from dataclasses import is_dataclass
from enum import Enum
from typing import (
    Any,
    Callable,
    Collection,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
)

from apischema import settings
from apischema.aliases import Aliaser
from apischema.cache import cache
from apischema.conversions.conversions import Conversions, to_hashable_conversions
from apischema.conversions.dataclass_models import DataclassModel
from apischema.conversions.visitor import SerializationVisitor
from apischema.dataclass_utils import (
    dataclass_types_and_fields,
    get_alias,
    get_field_conversion,
)
from apischema.fields import FIELDS_SET_ATTR, fields_set
from apischema.metadata.keys import SKIP_METADATA, check_metadata, is_aggregate_field
from apischema.serialization.serialized_methods import get_serialized_methods
from apischema.types import COLLECTION_TYPES, MAPPING_TYPES, PRIMITIVE_TYPES
from apischema.utils import OperationKind, Undefined
from apischema.visitor import Unsupported

PRIMITIVE_TYPES_SET = set(PRIMITIVE_TYPES)
COLLECTION_TYPE_SET = set(COLLECTION_TYPES)
MAPPING_TYPE_SET = set(MAPPING_TYPES)

SerializationMethod = Callable[[Any, bool, Callable], Any]


@cache
def serialization_fields(
    cls: Type, aliaser: Aliaser
) -> Tuple[
    Sequence[Tuple[str, str, SerializationMethod]],
    Sequence[Tuple[str, SerializationMethod]],
    Sequence[Tuple[str, SerializationMethod]],
]:
    types, fields, _ = dataclass_types_and_fields(cls)  # type: ignore
    normal_fields, aggregate_fields, serialized_fields = [], [], []
    for field in fields:
        if SKIP_METADATA in field.metadata:
            continue
        check_metadata(field)
        field_type, conversion = get_field_conversion(
            field, types[field.name], OperationKind.SERIALIZATION
        )
        method: Callable
        if conversion is not None:
            # method could be optimized when identity is used, but it's a lot more lines
            # for an unnecessary thing
            def method(
                obj: Any,
                exc_unset: bool,
                _serialize: Callable,
                conversions=conversion.conversions,  # type: ignore
                converter=conversion.converter,  # type: ignore
                exclude_unset=conversion.exclude_unset,  # type: ignore
            ) -> Any:
                if exclude_unset is not None:
                    exc_unset = exclude_unset
                return _serialize(converter(obj), exc_unset, conversions=conversions)

        elif field_type in PRIMITIVE_TYPES_SET:

            def method(obj: Any, exc_unset: bool, _serialize: Callable):
                return obj

        else:

            def method(obj: Any, exc_unset: bool, _serialize: Callable):
                return _serialize(obj, exc_unset)

        if is_aggregate_field(field):
            aggregate_fields.append((field.name, method))
        else:
            normal_fields.append((field.name, aliaser(get_alias(field)), method))
    for name, (serialized, _) in get_serialized_methods(cls).items():

        def method(
            obj: Any,
            exc_unset: bool,
            _serialize: Callable,
            func=serialized.func,
            conversions=serialized.conversions,
        ) -> Any:
            res = func(obj)
            return (
                res
                if res is Undefined
                else _serialize(res, exc_unset, conversions=conversions)
            )

        serialized_fields.append((aliaser(name), method))

    return normal_fields, aggregate_fields, serialized_fields


get_conversions = cache(SerializationVisitor.get_conversions)


def serialize(
    obj: Any,
    *,
    conversions: Conversions = None,
    aliaser: Aliaser = None,
    exclude_unset: bool = None,
) -> Any:
    if aliaser is None:
        aliaser = settings.aliaser()
    if exclude_unset is None:
        exclude_unset = settings.exclude_unset
    conversions = to_hashable_conversions(conversions)

    def _serialize(
        obj: Any,
        exc_unset: bool,
        *,
        conversions: Optional[Conversions] = None,
    ) -> Any:
        assert aliaser is not None
        cls = obj.__class__
        if cls in PRIMITIVE_TYPES_SET:
            return obj
        if cls in COLLECTION_TYPE_SET:
            return [_serialize(elt, exc_unset, conversions=conversions) for elt in obj]
        if cls in MAPPING_TYPE_SET:
            return {
                _serialize(key, exc_unset, conversions=conversions): _serialize(
                    value, exc_unset, conversions=conversions
                )
                for key, value in obj.items()
            }
        conversion = get_conversions(cls, conversions)
        if conversion is not None:
            if conversion.exclude_unset is not None:
                exc_unset = conversion.exclude_unset
            if isinstance(conversion.target, DataclassModel):
                cls = conversion.target.dataclass
            else:
                return _serialize(
                    conversion.converter(obj),  # type: ignore
                    exc_unset,
                    conversions=conversion.conversions,
                )
        if is_dataclass(cls):
            fields, aggregate_fields, serialized_fields = serialization_fields(
                cls, aliaser
            )
            if exclude_unset and hasattr(obj, FIELDS_SET_ATTR):
                fields_set_ = fields_set(obj)
                fields = [
                    (name, alias, method)
                    for (name, alias, method) in fields
                    if name in fields_set_
                ]
                aggregate_fields = [
                    (name, method)
                    for (name, method) in aggregate_fields
                    if name in fields_set_
                ]
            result = {}
            # properties before normal fields to avoid overloading a field with property
            for name, method in aggregate_fields:
                attr = getattr(obj, name)
                result.update(method(attr, exc_unset, _serialize))  # type: ignore
            for name, alias, method in fields:
                attr = getattr(obj, name)
                if attr is not Undefined:
                    result[alias] = method(attr, exc_unset, _serialize)  # type: ignore
            for alias, method in serialized_fields:
                res = method(obj, exc_unset, _serialize)
                if res is not Undefined:
                    result[alias] = res
            return result
        if obj is Undefined:
            raise Unsupported(cls)
        if issubclass(cls, Enum):
            return _serialize(obj.value, exc_unset)
        if isinstance(obj, PRIMITIVE_TYPES):
            return obj
        if isinstance(obj, Mapping):
            return {
                _serialize(key, exc_unset): _serialize(value, exc_unset)
                for key, value in obj.items()
            }
        if issubclass(cls, tuple) and hasattr(cls, "_fields"):
            result = {}
            for field_name in obj._fields:
                attr = getattr(obj, field_name)
                if attr is not Undefined:
                    result[aliaser(field_name)] = attr
            for name, (serialized, _) in get_serialized_methods(cls).items():
                res = serialized.func(obj)
                if res is not Undefined:
                    result[aliaser(name)] = _serialize(
                        res, exc_unset, conversions=serialized.conversions
                    )
            return result
        if isinstance(obj, Collection):
            return [_serialize(elt, exc_unset) for elt in obj]
        raise Unsupported(cls)

    return _serialize(obj, exclude_unset, conversions=conversions)



from collections import defaultdict
from typing import (
    Callable,
    Dict,
    List,
    Optional,
    TYPE_CHECKING,
    Type,
    TypeVar,
    overload,
)

from apischema.conversions.conversions import (
    Conversion,
    Conversions,
    resolve_deserialization,
    resolve_serialization,
)
from apischema.conversions.utils import converter_types
from apischema.utils import (
    MethodOrProperty,
    MethodWrapper,
    is_method,
    method_class,
    method_wrapper,
)

if TYPE_CHECKING:
    from apischema.deserialization.coercion import Coercion


_deserializers: Dict[Type, List[Conversion]] = defaultdict(list)
_serializers: Dict[Type, Conversion] = {}
Deserializer = TypeVar("Deserializer", Callable, Conversion, staticmethod, type)
Serializer = TypeVar("Serializer", Callable, Conversion, property, type)


class DeserializerDescriptor(MethodWrapper[staticmethod]):
    def __init__(self, method: staticmethod, **kwargs):
        super().__init__(method)
        self._kwargs = kwargs

    def __set_name__(self, owner, name):
        super().__set_name__(owner, name)
        method = self._method.__get__(None, object)
        source, target = converter_types(method, namespace={owner.__name__: owner})
        conversion = Conversion(method, source=source, target=target, **self._kwargs)
        resolved = resolve_deserialization(conversion)
        if resolved not in _deserializers[resolved.source]:
            _deserializers[resolved.target] = conversion


@overload
def deserializer(arg: Deserializer) -> Deserializer:
    ...


@overload
def deserializer(
    *,
    conversions: Conversions = None,
    additional_properties: Optional[bool] = None,
    coercion: Optional["Coercion"] = None,
    default_fallback: Optional[bool] = None,
) -> Callable[[Serializer], Serializer]:
    ...


def deserializer(arg=None, **kwargs):
    if arg is None:
        return lambda arg: deserializer(arg, **kwargs)  # type: ignore
    if isinstance(arg, staticmethod):
        return DeserializerDescriptor(arg, **kwargs)
    if kwargs and not isinstance(arg, Conversion):
        arg = Conversion(arg, **kwargs)
    resolved = resolve_deserialization(arg)
    if resolved not in _deserializers[resolved.target]:
        _deserializers[resolved.target].append(arg)
    return arg


class SerializerDescriptor(MethodWrapper[MethodOrProperty]):
    def __init__(self, method: MethodOrProperty, **kwargs):
        super().__init__(method)
        self._kwargs = kwargs

    def __set_name__(self, owner, name):
        super().__set_name__(owner, name)
        method = method_wrapper(self._method, name)
        source, target = converter_types(
            method, source=owner, namespace={owner.__name__: owner}
        )
        conversion = Conversion(method, source=source, target=target, **self._kwargs)
        _serializers[resolve_serialization(conversion).source] = conversion


@overload
def serializer(arg: Serializer) -> Serializer:
    ...


@overload
def serializer(
    *,
    conversions: Conversions = None,
    exclude_unset: bool = None,
) -> Callable[[Serializer], Serializer]:
    ...


def serializer(arg=None, **kwargs):
    if arg is None:
        return lambda arg: serializer(arg, **kwargs)  # type: ignore
    if is_method(arg):
        if method_class(arg) is None:
            return SerializerDescriptor(arg, **kwargs)
        else:
            arg = method_wrapper(arg)
    if kwargs and not isinstance(arg, Conversion):
        arg = Conversion(arg, **kwargs)
    _serializers[resolve_serialization(arg).source] = arg
    return arg


def reset_deserializers(cls: Type):
    _deserializers.pop(cls, ...)


def reset_serializer(cls: Type):
    _deserializers.pop(cls, ...)


class InheritedDeserializer:
    def __init__(self, method: classmethod, **kwargs):
        self.method = method
        self.kwargs = kwargs

    def __set_name__(self, owner, name):
        prev_init_subclass = owner.__init_subclass__

        def init_subclass(cls, **kwargs):
            prev_init_subclass(**kwargs)
            method = self.method.__get__(None, cls)
            deserializer(Conversion(method, target=cls, **self.kwargs))

        owner.__init_subclass__ = classmethod(init_subclass)
        init_subclass(owner)
        setattr(owner, name, self.method)


ClsMethod = TypeVar("ClsMethod")


@overload
def inherited_deserializer(method: ClsMethod) -> ClsMethod:
    ...


@overload
def inherited_deserializer(
    *,
    conversions: Conversions = None,
    additional_properties: Optional[bool] = None,
    coercion: Optional["Coercion"] = None,
    default_fallback: Optional[bool] = None,
) -> Callable[[ClsMethod], ClsMethod]:
    ...


def inherited_deserializer(method=None, **kwargs):
    if method is None:
        return lambda func: inherited_deserializer(func, **kwargs)  # type: ignore
    if not isinstance(method, classmethod):
        raise TypeError("inherited_deserializer must be called on classmethod")
    return InheritedDeserializer(method, **kwargs)


Cls = TypeVar("Cls", bound=Type)


def as_str(cls: Cls) -> Cls:
    deserializer(Conversion(cls, source=str))
    serializer(Conversion(str, source=cls))
    return cls



from typing import Generic, Optional, Sequence, TypeVar

from dataclasses import dataclass, field

from apischema.graphql import resolver
from apischema.metadata import skip

Cursor_ = TypeVar("Cursor_")


@dataclass
class PageInfo(Generic[Cursor_]):
    has_previous_page: bool
    has_next_page: bool
    start_cursor: Optional[Cursor_]
    end_cursor: Optional[Cursor_]


Node_ = TypeVar("Node_")


@dataclass
class Edge(Generic[Node_, Cursor_]):
    node: Node_
    cursor: Cursor_


Edge_ = TypeVar("Edge_", bound=Edge)


@dataclass
class Connection(Generic[Node_, Cursor_, Edge_]):
    edges: Optional[Sequence[Optional[Edge_]]]
    has_previous_page: bool = field(default=False, metadata=skip)
    has_next_page: bool = field(default=False, metadata=skip)
    start_cursor: Optional[Cursor_] = field(default=None, metadata=skip)
    end_cursor: Optional[Cursor_] = field(default=None, metadata=skip)

    def page_info(self) -> PageInfo[Cursor_]:
        start_cursor, end_cursor = None, None
        if self.edges is not None:
            if self.edges[0] is not None:
                start_cursor = self.edges[0].cursor
            if self.edges[-1] is not None:
                end_cursor = self.edges[-1].cursor
        if self.start_cursor is not None:
            start_cursor = self.start_cursor
        if self.end_cursor is not None:
            end_cursor = self.end_cursor
        return PageInfo(
            has_previous_page=self.has_previous_page,
            has_next_page=self.has_next_page,
            start_cursor=start_cursor,
            end_cursor=end_cursor,
        )


resolver(Connection.page_info)



import operator
import re
from dataclasses import dataclass, field, fields
from math import gcd
from typing import (
    Any,
    Callable,
    ClassVar,
    Collection,
    Dict,
    List,
    Mapping,
    Optional,
    Pattern,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from apischema.types import Number
from apischema.utils import merge_opts, to_hashable
from apischema.validation.errors import ValidationError

VALIDATION_METADATA = "validation"
MODIFIER_METADATA = "modifier"
MERGE_METADATA = "merge"
ALIAS_METADATA = "alias"

T = TypeVar("T")


def comparison_field(comparison: Callable, alias: str = None):
    by_comp = {
        operator.lt: (max, "less than"),
        operator.le: (max, "less than or equal to"),
        operator.gt: (min, "greater than"),
        operator.ge: (min, "greater than or equal to"),
    }
    merge, err = by_comp[comparison]
    metadata = {
        VALIDATION_METADATA: (comparison, err + " {}"),
        MODIFIER_METADATA: ...,
        MERGE_METADATA: merge,
    }
    if alias is not None:
        metadata[ALIAS_METADATA] = alias
    return field(default=None, metadata=metadata)


_Constraints = TypeVar("_Constraints", bound="Constraints")


@dataclass(frozen=True)
class Constraints:
    valid_types: ClassVar[Union[Type, Tuple[Type, ...]]]
    collection_prefix: ClassVar[Optional[str]] = None

    def __post_init__(self):
        checks: List[Tuple[Any, Callable, Callable, str]] = []
        for f in fields(self):
            attr = getattr(self, f.name)
            if attr is None:
                continue
            comp, err = f.metadata[VALIDATION_METADATA]
            error = err.format(attr) + f" ({f.metadata.get(ALIAS_METADATA, f.name)})"
            modif = f.metadata.get(MODIFIER_METADATA)
            if modif is ...:
                if self.collection_prefix is not None:
                    error = f"{self.collection_prefix} {error}"
                    modif = len
                else:
                    modif = None
            if modif is None:
                modif = lambda x: x  # noqa: E731
            checks.append((attr, comp, modif, error))

        def errors(data: Any) -> List[str]:
            assert isinstance(data, self.valid_types)
            return [err for attr, comp, modif, err in checks if comp(modif(data), attr)]

        object.__setattr__(self, self.errors.__name__, errors)

    def validate(self, data: T) -> T:
        errors = self.errors(data)
        if errors:
            raise ValidationError(errors)
        return data

    def errors(self, data: Any) -> List[str]:
        raise NotImplementedError  # initialized in __post_init__

    def as_dict(self) -> Mapping[str, Any]:
        return {
            f.metadata.get(ALIAS_METADATA, f.name): getattr(self, f.name)
            for f in fields(self)
            if getattr(self, f.name) is not None
        }


Cls = TypeVar("Cls", bound=type)


@dataclass(frozen=True)
class NumberConstraints(Constraints):
    valid_types = (int, float)
    min: Optional[Number] = comparison_field(operator.lt, "minimum")
    max: Optional[Number] = comparison_field(operator.gt, "maximum")
    exc_min: Optional[Number] = comparison_field(operator.le, "exclusiveMinimum")
    exc_max: Optional[Number] = comparison_field(operator.ge, "exclusiveMaximum")
    mult_of: Optional[Number] = field(
        default=None,
        metadata={
            VALIDATION_METADATA: (operator.mod, "not a multiple of {}"),
            MERGE_METADATA: lambda m1, m2: m1 * m2 / gcd(m1, m2),
            ALIAS_METADATA: "multipleOf",
        },
    )


def merge_pattern(p1: Pattern, p2: Pattern) -> Pattern:
    raise TypeError("Cannot merge patterns")


class PatternNotMatched(str):
    def format(self, arg: Pattern) -> str:  # type: ignore
        return f"'{arg.pattern}' not matched"


@dataclass(frozen=True)
class StringConstraints(Constraints):
    valid_types = str
    collection_prefix = "length"
    min_len: Optional[int] = comparison_field(operator.lt, "minLength")
    max_len: Optional[int] = comparison_field(operator.gt, "maxLength")
    pattern: Optional[Union[str, Pattern]] = field(
        default=None,
        metadata={
            VALIDATION_METADATA: (lambda d, p: not p.match(d), PatternNotMatched()),
            MERGE_METADATA: merge_pattern,
        },
    )

    def __post_init__(self):
        if self.pattern is not None:
            object.__setattr__(self, "pattern", re.compile(self.pattern))
        super().__post_init__()


@dataclass(frozen=True)
class ArrayConstraints(Constraints):
    valid_types = Collection
    collection_prefix = "size"
    min_items: Optional[int] = comparison_field(operator.lt, "minItems")
    max_items: Optional[int] = comparison_field(operator.gt, "maxItems")
    unique: Optional[bool] = field(
        default=None,
        metadata={
            VALIDATION_METADATA: (operator.ne, "duplicate items"),
            MODIFIER_METADATA: lambda d: len(set(map(to_hashable, d))) == len(d),
            MERGE_METADATA: operator.or_,
            ALIAS_METADATA: "uniqueItems",
        },
    )


@dataclass(frozen=True)
class ObjectConstraints(Constraints):
    valid_types = Mapping
    collection_prefix = "size"
    min_props: Optional[int] = comparison_field(operator.lt, "minProperties")
    max_props: Optional[int] = comparison_field(operator.gt, "maxProperties")


@merge_opts
def merge_constraints(c1: Constraints, c2: Constraints) -> Constraints:
    if type(c2) != type(c1):
        raise TypeError("Incompatible constraints types")
    constraints: Dict[str, Any] = {}
    for field_ in fields(c1):
        if getattr(c1, field_.name) is None:
            constraints[field_.name] = getattr(c2, field_.name)
        elif getattr(c2, field_.name) is None:
            constraints[field_.name] = getattr(c1, field_.name)
        else:
            constraints[field_.name] = field_.metadata[MERGE_METADATA](
                getattr(c1, field_.name), getattr(c2, field_.name)
            )
    return type(c1)(**constraints)  # type: ignore



from collections import defaultdict
from dataclasses import Field
from itertools import chain
from typing import AbstractSet, Any, Collection, Mapping, Set, overload

from apischema.utils import PREFIX

DEPENDENT_REQUIRED_ATTR = f"{PREFIX}dependent_required"

Requirements = Mapping[Field, AbstractSet[Field]]


class DependentRequired:
    @overload
    def __init__(self, fields: Mapping[Any, Collection[Any]]):
        ...

    @overload
    def __init__(self, fields: Mapping[Any, Collection[Any]], *groups: Collection[Any]):
        ...

    @overload
    def __init__(self, *groups: Collection[Any]):
        ...

    def __init__(self, *groups: Collection[Any]):  # type: ignore
        from apischema.metadata.keys import (
            MERGED_METADATA,
            PROPERTIES_METADATA,
            SKIP_METADATA,
            is_aggregate_field,
        )

        if not groups:
            return
        fields: Mapping[Any, Collection[Any]] = {}
        if isinstance(groups[0], Mapping):
            fields, *groups = groups  # type: ignore
        for field in chain(*groups, fields, chain.from_iterable(fields.values())):
            if not isinstance(field, Field):
                raise TypeError("Dependency must be a field")
            if is_aggregate_field(field):
                raise TypeError("Dependency cannot be a aggregate field")
            if SKIP_METADATA in field.metadata:
                raise TypeError("Cannot use skipped field as dependency")
            if (
                MERGED_METADATA in field.metadata
                or PROPERTIES_METADATA in field.metadata
            ):
                raise TypeError("Cannot use aggregate field as dependency")
        self.fields: Mapping[Field, Collection[Field]] = fields
        self.groups: Collection[Collection[Field]] = groups

    def __set_name__(self, owner, name):
        others = getattr(owner, DEPENDENT_REQUIRED_ATTR, ())
        setattr(owner, DEPENDENT_REQUIRED_ATTR, (*others, self))

    def _group_requirements(self) -> Mapping[Field, Set[Field]]:
        result: Mapping[Field, Set[Field]] = defaultdict(set)
        for group in self.groups:
            group = set(group)
            for field in group:
                result[field].update(group - {field})
        return result

    def required_by(self) -> Requirements:
        result = self._group_requirements()
        for field, required in self.fields.items():
            for req in required:
                result[req].add(field)
        return result

    def requiring(self) -> Requirements:
        result = self._group_requirements()
        for field, required in self.fields.items():
            for req in required:
                result[field].add(req)
        return result



from functools import lru_cache
from typing import Any, Callable, List, TypeVar, cast

_cached: List = []

Func = TypeVar("Func", bound=Callable)


def cache(func: Func) -> Func:
    cached: Any = lru_cache()(func)
    _cached.append(cached)
    return cast(Func, cached)


def reset_cache():
    for cached in _cached:
        cached.cache_clear()



__all__ = [
    "ID",
    "Mutation",
    "Query",
    "Subscription",
    "graphql_schema",
    "interface",
    "relay",
    "resolver",
]


try:
    from .schema import ID, Query, Mutation, Subscription, graphql_schema
    from .interfaces import interface
    from .resolvers import resolver
    from . import relay
except ImportError:
    raise ImportError(
        "GraphQL feature requires graphql-core library\n"
        "Run `pip install apischema[graphql]` to install it"
    )



import collections.abc
import re
import sys
from dataclasses import Field, dataclass, field
from functools import reduce, wraps
from inspect import isgeneratorfunction
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    cast,
    overload,
)

from apischema.aliases import Aliaser
from apischema.dataclass_utils import get_alias
from apischema.dataclasses import replace
from apischema.typing import get_args, get_origin
from apischema.utils import get_args2, get_origin2, merge_opts

try:
    from apischema.typing import Annotated
except ImportError:
    Annotated = ...  # type: ignore

ErrorMsg = str
Error = Union[ErrorMsg, Tuple[Any, ErrorMsg]]
# where  Any = Union[Field, int, str, Iterable[Union[Field, int, str,]]]
# but Field being kind of magic not understood by type checkers, it's hidden behind Any
ErrorKey = Union[str, int]
T = TypeVar("T")
ValidatorResult = Generator[Error, None, T]


@dataclass
class LocalizedError:
    loc: Sequence[ErrorKey]
    err: Sequence[ErrorMsg]

    def nested(self, index=0) -> "ValidationError":
        if index == len(self.loc):
            return ValidationError(self.err)
        else:
            assert index < len(self.loc)
            return ValidationError(children={self.loc[index]: self.nested(index + 1)})


@dataclass
class ValidationError(Exception):
    messages: Sequence[ErrorMsg] = field(default_factory=list)
    children: Mapping[ErrorKey, "ValidationError"] = field(default_factory=dict)

    def flat(self) -> Iterator[Tuple[Tuple[ErrorKey, ...], Sequence[ErrorMsg]]]:
        if self.messages:
            yield (), self.messages
        for child_key in sorted(self.children):
            for path, errors in self.children[child_key].flat():
                yield (child_key, *path), errors

    def serialize(self) -> Sequence[LocalizedError]:
        return [LocalizedError(loc, err) for loc, err in self.flat()]

    @staticmethod
    def deserialize(errors: Sequence[LocalizedError]) -> "ValidationError":
        return reduce(
            merge_errors, map(LocalizedError.nested, errors), ValidationError()
        )


@overload
def merge_errors(
    err1: Optional[ValidationError], err2: ValidationError
) -> ValidationError:
    ...


@overload
def merge_errors(
    err1: ValidationError, err2: Optional[ValidationError]
) -> ValidationError:
    ...


@overload
def merge_errors(
    err1: Optional[ValidationError], err2: Optional[ValidationError]
) -> Optional[ValidationError]:
    ...


@merge_opts  # type: ignore
def merge_errors(err1: ValidationError, err2: ValidationError) -> ValidationError:
    if err1 is None:
        return err2
    if err2 is None:
        return err1
    return ValidationError(
        [*err1.messages, *err2.messages],
        {
            key: merge_errors(  # type: ignore
                err1.children.get(key), err2.children.get(key)
            )
            for key in err1.children.keys() | err2.children.keys()
        },
    )


def exception(err: Exception) -> str:
    return str(err)


def _rec_build_error(path: Sequence[str], msg: ErrorMsg) -> ValidationError:
    if not path:
        return ValidationError([msg])
    else:
        return ValidationError(children={path[0]: _rec_build_error(path[1:], msg)})


class FieldPath(str):
    def __new__(cls, field: Field):
        obj = super().__new__(cls, field.name)  # type: ignore
        obj.field = field  # type: ignore
        return obj


def _check_error_path(path) -> Sequence[ErrorKey]:
    if isinstance(path, (Field, int, str)):
        path = [path]
    else:
        path = list(path)
    for i, elt in enumerate(path):
        if isinstance(elt, Field):
            path[i] = FieldPath(elt)
        if not isinstance(path[i], (str, int)):
            raise TypeError(
                f"Bad error path, expected Field, int or str," f" found {type(i)}"
            )
    return cast(Sequence[ErrorKey], path)


def build_validation_error(errors: Iterable[Error]) -> ValidationError:
    messages: List[ErrorMsg] = []
    children: Dict[ErrorKey, ValidationError] = {}
    for error in errors:
        if isinstance(error, ErrorMsg):
            messages.append(error)
            continue
        path, msg = error
        path = _check_error_path(path)
        if not path:
            messages.append(msg)
        else:
            key, *remain = path
            children[key] = merge_errors(
                children.get(key), _rec_build_error(remain, msg)
            )
    return ValidationError(messages, children)


def apply_aliaser(error: ValidationError, aliaser: Aliaser) -> ValidationError:
    path_replace: Dict[ErrorKey, str] = {}
    error_replace = {}
    for path, child_error in error.children.items():
        if isinstance(path, FieldPath):
            path_replace[path] = aliaser(get_alias(path.field))  # type: ignore
        new_error = apply_aliaser(child_error, aliaser)
        if new_error is not child_error:
            error_replace[path] = new_error
    if path_replace or error_replace:
        return replace(
            error,
            children={
                path_replace.get(path, path): error_replace.get(path, err)
                for path, err in error.children.items()
            },
        )
    else:
        return error


if sys.version_info >= (3, 7):  # pragma: no cover
    GeneratorOrigin = collections.abc.Generator
else:
    GeneratorOrigin = Generator  # pragma: no cover


def yield_to_raise(func: Callable[..., ValidatorResult[T]]) -> Callable[..., T]:
    @wraps(func)
    def wrapper(*args, **kwargs):
        result, errors = func(*args, **kwargs), []
        while True:
            try:
                errors.append(next(result))
            except StopIteration as stop:
                if errors:
                    raise build_validation_error(errors)
                return stop.value

    return wrapper


# Looking forward to PEP 612
def with_validation_error(func: Callable[..., ValidatorResult[T]]) -> Callable[..., T]:
    if not isgeneratorfunction(func):
        raise TypeError("func must be a generator returning a ValidatorResult")
    wrapper = yield_to_raise(func)
    if "return" in func.__annotations__:
        ret = func.__annotations__["return"]
        if isinstance(ret, str):
            match = re.match(r"ValidatorResult\[(?P<ret>.*)\]", ret)
            if match is not None:
                ret = match.groupdict("ret")
        else:
            annotations = get_args(ret)[1:] if get_origin(ret) == Annotated else ()
            if get_origin2(ret) == GeneratorOrigin:
                ret = get_args2(ret)[2]
                if annotations:
                    ret = Annotated[(ret, *annotations)]
        wrapper.__annotations__["return"] = ret
    return wrapper



# flake8: noqa
from dataclasses import *


def _replace(__obj, **changes):
    from apischema.fields import FIELDS_SET_ATTR, fields_set, set_fields
    from dataclasses import replace as replace_, _FIELDS, _FIELD_INITVAR  # type: ignore

    # Fix https://bugs.python.org/issue36470
    assert is_dataclass(__obj)
    for name, field in getattr(__obj, _FIELDS).items():
        if field._field_type == _FIELD_INITVAR and name not in changes:  # type: ignore
            if field.default is not MISSING:
                changes[name] = field.default
            elif field.default_factory is not MISSING:
                changes[name] = field.default_factory()

    result = replace_(__obj, **changes)
    if hasattr(__obj, FIELDS_SET_ATTR):
        set_fields(result, *fields_set(__obj), *changes, overwrite=True)
    return result


globals()[replace.__name__] = _replace

del _replace



"""Kind of typing_extensions for this package"""
__all__ = ["_LiteralMeta", "_TypedDictMeta", "get_args", "get_origin", "get_type_hints"]

import sys
from types import ModuleType
from typing import Any, Callable, Generic, Tuple, TypeVar, _eval_type  # type: ignore


class _FakeType:
    pass


if sys.version_info >= (3, 9):  # pragma: no cover
    from typing import Annotated, get_type_hints, get_origin, get_args
else:  # pragma: no cover
    try:
        from typing_extensions import Annotated
    except ImportError:
        pass
    try:
        from typing_extensions import get_type_hints as gth
    except ImportError:
        from typing import get_type_hints as _gth

        def gth(obj, globalns=None, localns=None, include_extras=False):  # type: ignore
            return _gth(obj, globalns, localns)

    def get_type_hints(  # type: ignore
        obj, globalns=None, localns=None, include_extras=False
    ):
        # TODO This has been fixed in recent 3.7 and 3.8
        # fix https://bugs.python.org/issue37838
        if not isinstance(obj, (type, ModuleType)) and globalns is None:
            nsobj = obj
            while hasattr(nsobj, "__wrapped__"):
                nsobj = nsobj.__wrapped__
            globalns = getattr(nsobj, "__globals__", None)
        localns = {"unicode": str, **(localns or {})}
        return gth(obj, globalns, localns, include_extras)

    try:
        from typing_extensions import get_origin, get_args
    except ImportError:

        def _assemble_tree(tree: Tuple[Any]) -> Any:
            if not isinstance(tree, tuple):
                return tree
            else:
                origin, *args = tree  # type: ignore
                if origin is Annotated:
                    return Annotated[(_assemble_tree(args[0]), *args[1])]
                else:
                    return origin[tuple(map(_assemble_tree, args))]

        def get_origin(tp):  # type: ignore
            # In Python 3.6: List[Collection[T]][int]._args__ == int != Collection[int]
            if hasattr(tp, "_subs_tree"):
                tp = _assemble_tree(tp._subs_tree())
            if isinstance(tp, _AnnotatedAlias):
                return Annotated
            if tp is Generic:
                return Generic
            return getattr(tp, "__origin__", None)

        def get_args(tp):  # type: ignore
            # In Python 3.6: List[Collection[T]][int]._args__ == int != Collection[int]
            if hasattr(tp, "_subs_tree"):
                tp = _assemble_tree(tp._subs_tree())
            if isinstance(tp, _AnnotatedAlias):
                return (tp.__args__[0], *tp.__metadata__)
            # __args__ can be None in 3.6 inside __set_name__
            res = getattr(tp, "__args__", ()) or ()
            if get_origin(tp) is Callable and res[0] is not Ellipsis:
                res = (list(res[:-1]), res[-1])
            return res


if sys.version_info >= (3, 7):
    from typing import _collect_type_vars, ForwardRef  # type: ignore
else:
    from typing import _type_vars, _ForwardRef

    _collect_type_vars = _type_vars

    def ForwardRef(arg, is_argument):
        return _ForwardRef(arg)


try:
    from typing import _strip_annotations  # type: ignore
except ImportError:
    try:
        from typing_extensions import _strip_annotations  # type: ignore
    except ImportError:

        def _strip_annotations(t):
            return t


def _generic_mro(result, tp):
    origin = get_origin(tp)
    if origin is None:
        origin = tp
    result[origin] = tp
    if hasattr(origin, "__orig_bases__"):
        parameters = _collect_type_vars(origin.__orig_bases__)
        substitution = dict(zip(parameters, get_args(tp)))
        for base in origin.__orig_bases__:
            if get_origin(base) in result:
                continue
            base_parameters = getattr(base, "__parameters__", ())
            if base_parameters:
                base = base[tuple(substitution.get(p, p) for p in base_parameters)]
            _generic_mro(result, base)


def generic_mro(tp):
    origin = get_origin(tp)
    if origin is None and not hasattr(tp, "__orig_bases__"):
        if not isinstance(tp, type):
            raise TypeError(f"{tp!r} is not a type or a generic alias")
        return tp.__mro__
    # sentinel value to avoid to subscript Generic and Protocol
    try:
        result = {Generic: Generic, Protocol: Protocol}
    except NameError:
        result = {Generic: Generic}
    _generic_mro(result, tp)
    cls = origin if origin is not None else tp
    return tuple(result.get(sub_cls, sub_cls) for sub_cls in cls.__mro__)


def _class_annotations(cls, globalns, localns):
    hints = {}
    if globalns is None:
        base_globals = sys.modules[cls.__module__].__dict__
    else:
        base_globals = globalns
    for name, value in cls.__dict__.get("__annotations__", {}).items():
        if value is None:
            value = type(None)
        if isinstance(value, str):
            value = ForwardRef(value, is_argument=False)
        hints[name] = _eval_type(value, base_globals, localns)
    return hints


def get_type_hints2(obj, globalns=None, localns=None):  # type: ignore
    if isinstance(obj, type) or isinstance(get_origin(obj), type):
        hints = {}
        for base in reversed(generic_mro(obj)):
            origin = get_origin(base)
            if hasattr(origin, "__orig_bases__"):
                parameters = _collect_type_vars(origin.__orig_bases__)
                substitution = dict(zip(parameters, get_args(base)))
                annotations = _class_annotations(get_origin(base), globalns, localns)
                for name, tp in annotations.items():
                    if isinstance(tp, TypeVar):
                        hints[name] = substitution.get(tp, tp)
                    elif getattr(tp, "__parameters__", ()):
                        hints[name] = tp[
                            tuple(substitution.get(p, p) for p in tp.__parameters__)
                        ]
                    else:
                        hints[name] = tp
            else:
                hints.update(_class_annotations(base, globalns, localns))
        return hints
    else:
        return get_type_hints(obj, globalns, localns, include_extras=True)


if sys.version_info >= (3, 8):  # pragma: no cover
    from typing import Literal, TypedDict, Protocol  # noqa: F401
else:  # pragma: no cover
    try:
        from typing_extensions import Literal, TypedDict, Protocol  # noqa: F401
    except ImportError:
        pass


_T = TypeVar("_T")
_GenericAlias: Any = type(Generic[_T])
try:
    _AnnotatedAlias: Any = type(Annotated[_T, ...])
except NameError:
    _AnnotatedAlias = _FakeType
try:

    class _TypedDictImplem(TypedDict):
        pass

    _LiteralMeta: Any = type(Literal)
    _TypedDictMeta: Any = type(_TypedDictImplem)
except NameError:
    _LiteralMeta, _TypedDictMeta = _FakeType, _FakeType  # type: ignore



import collections.abc
import re
from enum import Enum, auto
from functools import wraps
from types import FunctionType
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Hashable,
    Iterable,
    List,
    Mapping,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

from apischema.types import (
    AnyType,
    COLLECTION_TYPES,
    MAPPING_TYPES,
    subscriptable_origin,
)
from apischema.typing import _GenericAlias, _collect_type_vars, get_args, get_origin

try:
    from apischema.typing import Annotated
except ImportError:
    Annotated = ...  # type: ignore

PREFIX = "_apischema_"

T = TypeVar("T")
U = TypeVar("U")


# Singleton type, see https://www.python.org/dev/peps/pep-0484/#id30
class UndefinedType(Enum):
    def __repr__(self):
        return "Undefined"

    def __str__(self):
        return "Undefined"

    def __bool__(self):
        return False

    Undefined = auto()


Undefined = UndefinedType.Undefined


def opt_or(opt: Optional[T], default: U) -> Union[T, U]:
    return opt if opt is not None else default


def to_hashable(data: Union[None, int, float, str, bool, list, dict]) -> Hashable:
    if isinstance(data, list):
        return tuple(map(to_hashable, data))
    if isinstance(data, dict):
        return tuple(sorted((to_hashable(k), to_hashable(v)) for k, v in data.items()))
    return data  # type: ignore


SNAKE_CASE_REGEX = re.compile(r"_([a-z\d])")


def to_camel_case(s: str):
    return SNAKE_CASE_REGEX.sub(lambda m: m.group(1).upper(), s)


def type_name(tp: AnyType) -> str:
    if hasattr(tp, "__name__"):
        return tp.__name__
    elif isinstance(tp, _GenericAlias):
        return tp._name
    else:
        raise NotImplementedError


MakeDataclassField = Union[Tuple[str, AnyType], Tuple[str, AnyType, Any]]


def merge_opts(
    func: Callable[[T, T], T]
) -> Callable[[Optional[T], Optional[T]], Optional[T]]:
    def wrapper(opt1, opt2):
        if opt1 is None:
            return opt2
        if opt2 is None:
            return opt1
        return func(opt1, opt2)

    return wrapper


K = TypeVar("K")
V = TypeVar("V")


@merge_opts
def merge_opts_mapping(m1: Mapping[K, V], m2: Mapping[K, V]) -> Mapping[K, V]:
    return {**m1, **m2}


def is_type_var(tp: AnyType) -> bool:
    return isinstance(tp, TypeVar)  # type: ignore


def has_type_vars(tp: AnyType) -> bool:
    return is_type_var(tp) or bool(getattr(tp, "__parameters__", ()))


TV = AnyType  # TypeVar is not supported as a type
# 10 should be enough for all builtin types
_type_vars = [TypeVar(f"T{i}") for i in range(10)]


def get_parameters(tp: AnyType) -> Iterable[TV]:
    if hasattr(tp, "__parameters__"):
        return tp.__parameters__
    elif hasattr(tp, "__orig_bases__"):
        return _collect_type_vars(tp.__orig_bases__)
    else:
        return _type_vars


def substitute_type_vars(tp: AnyType, substitution: Mapping[TV, AnyType]) -> AnyType:
    if is_type_var(tp):
        try:
            return substitution[tp]
        except KeyError:
            return Union[tp.__constraints__] if tp.__constraints__ else Any
    elif getattr(tp, "__parameters__", ()):
        return tp[tuple(substitution.get(p, p) for p in tp.__parameters__)]
    else:
        return tp


Func = TypeVar("Func", bound=Callable)


def typed_wraps(wrapped: Func) -> Callable[[Callable], Func]:
    return cast(Func, wraps(wrapped))


def get_origin2(tp: AnyType) -> Optional[Type]:
    origin = get_origin(tp)
    return get_origin(get_args(tp)[0]) if origin is Annotated else origin


def get_args2(tp: AnyType) -> Tuple[AnyType, ...]:
    origin = get_origin(tp)
    return get_args(get_args(tp)[0]) if origin is Annotated else get_args(tp)


def get_origin_or_type(tp: AnyType) -> Type:
    origin = get_origin2(tp)
    return origin if origin is not None else tp


def is_union_of(tp: AnyType, of: AnyType) -> bool:
    return tp == of or (get_origin_or_type(tp) == Union and of in get_args2(tp))


class OperationKind(Enum):
    DESERIALIZATION = auto()
    SERIALIZATION = auto()


MethodOrProperty = Union[Callable, property]


def _method_location(method: MethodOrProperty) -> Optional[Type]:
    if isinstance(method, property):
        method = method.fget
    while hasattr(method, "__wrapped__"):
        method = method.__wrapped__  # type: ignore
    assert isinstance(method, FunctionType)
    global_name, *class_path = method.__qualname__.split(".")[:-1]
    if global_name not in method.__globals__:
        return None
    location = method.__globals__[global_name]
    for attr in class_path:
        if hasattr(location, attr):
            location = getattr(location, attr)
        else:
            break
    return location


def is_method(method: MethodOrProperty) -> bool:
    """Return if the function is method/property declared in a class"""
    return (isinstance(method, property) and is_method(method.fget)) or (
        isinstance(method, FunctionType)
        and method.__name__ != method.__qualname__
        and isinstance(_method_location(method), (type, type(None)))
    )


def method_class(method: MethodOrProperty) -> Optional[Type]:
    cls = _method_location(method)
    return cls if isinstance(cls, type) else None


METHOD_WRAPPER_ATTR = f"{PREFIX}method_wrapper"


def method_wrapper(method: MethodOrProperty, name: str = None) -> Callable:
    if isinstance(method, property):
        name = name or method.fget.__name__

        @wraps(method.fget)
        def wrapper(self):
            return getattr(self, name)

    else:
        if hasattr(method, METHOD_WRAPPER_ATTR):
            return method
        name = name or method.__name__

        @wraps(method)
        def wrapper(self, *args, **kwargs):
            return getattr(self, name)(*args, **kwargs)

    setattr(wrapper, METHOD_WRAPPER_ATTR, True)
    return wrapper


class MethodWrapper(Generic[T]):
    def __init__(self, method: T):
        self._method = method

    def getter(self, func):
        self._method.getter(func)
        return self

    def setter(self, func):
        self._method.setter(func)
        return self

    def deleter(self, func):
        self._method.deleter(func)
        return self

    def __set_name__(self, owner, name):
        setattr(owner, name, self._method)


def replace_builtins(tp: AnyType) -> AnyType:
    origin = get_origin_or_type(tp)
    args = tuple(map(replace_builtins, get_args2(tp)))
    if origin in COLLECTION_TYPES:
        if issubclass(origin, collections.abc.Set):
            replacement = subscriptable_origin(Set[None])
        else:
            replacement = subscriptable_origin(List[None])
    elif origin in MAPPING_TYPES:
        replacement = subscriptable_origin(Dict[None, None])
    else:
        replacement = origin
    res = replacement[args] if args else replacement
    return Annotated[(res, *get_args(tp)[1:])] if get_origin(tp) == Annotated else res


try:
    from functools import cached_property
except ImportError:
    # From 3.9 functools
    from threading import RLock

    _NOT_FOUND = object()

    class cached_property:  # type: ignore
        def __init__(self, func):
            self.func = func
            self.attrname = None
            self.__doc__ = func.__doc__
            self.lock = RLock()

        def __set_name__(self, owner, name):
            if self.attrname is None:
                self.attrname = name
            elif name != self.attrname:
                raise TypeError(
                    "Cannot assign the same cached_property to two different names "
                    f"({self.attrname!r} and {name!r})."
                )

        def __get__(self, instance, owner=None):
            if instance is None:
                return self
            if self.attrname is None:
                raise TypeError(
                    "Cannot use cached_property instance"
                    " without calling __set_name__ on it."
                )
            try:
                cache = instance.__dict__
            except AttributeError:  # not all objects have __dict__ (e.g. class defines slots) # noqa: E501
                msg = (
                    f"No '__dict__' attribute on {type(instance).__name__!r} "
                    f"instance to cache {self.attrname!r} property."
                )
                raise TypeError(msg) from None
            val = cache.get(self.attrname, _NOT_FOUND)
            if val is _NOT_FOUND:
                with self.lock:
                    # check if another thread filled cache while we awaited lock
                    val = cache.get(self.attrname, _NOT_FOUND)
                    if val is _NOT_FOUND:
                        val = self.func(instance)
                        try:
                            cache[self.attrname] = val
                        except TypeError:
                            msg = (
                                f"The '__dict__' attribute on"
                                f" {type(instance).__name__!r} instance "
                                f"does not support item assignment for"
                                f" caching {self.attrname!r} property."
                            )
                            raise TypeError(msg) from None
            return val

        __class_getitem__ = classmethod(_GenericAlias)



from dataclasses import dataclass, fields
from functools import partial
from types import FunctionType, MethodType
from typing import Any, Mapping, Optional, TYPE_CHECKING, Type, TypeVar

from apischema.dataclass_utils import get_default, has_default
from apischema.fields import FIELDS_SET_ATTR, set_fields

if TYPE_CHECKING:
    from apischema.validation.validator import Validator

MOCK_FIELDS_FIELD = "__mock_fields__"
MOCK_CLS_FIELD = "__mock_cls__"


class NonTrivialDependency(Exception):
    def __init__(self, attr: str):
        self.attr = attr
        self.validator: Optional["Validator"] = None


@dataclass(init=False)
class ValidatorMock:
    def __init__(self, cls: Type, values: Mapping[str, Any]):
        self.cls = cls
        self.values = values
        set_fields(self, *values, overwrite=True)

    def __getattribute__(self, name: str) -> Any:
        values = super().__getattribute__("values")
        if name in values:
            return values[name]
        cls = super().__getattribute__("cls")
        for field in fields(cls):
            if name == field.name:
                try:
                    return get_default(field)
                except NotImplementedError:
                    raise NonTrivialDependency(name) from None
        if name == "__class__":
            return cls
        if name == "__dict__":
            return {
                **values,
                **{
                    field.name: get_default(field)
                    for field in fields(cls)
                    if has_default(field)
                },
                FIELDS_SET_ATTR: set(values),
            }
        if name == FIELDS_SET_ATTR:
            return set(values)
        if hasattr(cls, name):
            member = getattr(cls, name)
            # for classmethod (staticmethod are not handled)
            if isinstance(member, MethodType):
                return member
            if isinstance(member, FunctionType):
                return partial(member, self)
            if isinstance(member, property):
                return member.fget(self)  # type: ignore
            return member
        raise NonTrivialDependency(name)


T = TypeVar("T")



import sys
from abc import ABC, abstractmethod
from typing import (
    Awaitable,
    ClassVar,
    Collection,
    Dict,
    Generic,
    List,
    Type,
    TypeVar,
    Union,
    cast,
)

import graphql
from dataclasses import Field, _FIELDS, dataclass, field  # type: ignore

from apischema import deserialize, deserializer, serialize, serializer
from apischema.graphql import interface
from apischema.graphql.schema import ID
from apischema.json_schema.refs import get_ref
from apischema.metadata import conversion
from apischema.types import AnyType
from apischema.typing import generic_mro, get_args, get_origin
from apischema.utils import has_type_vars

Id = TypeVar("Id")


class InvalidGlobalId(Exception):
    def __init__(self, value: str):
        self.value = value

    def __str__(self):
        return f"{self.value} is not a valid id"


class NotANode(Exception):
    def __init__(self, node_class: str):
        self.node_class = node_class

    def __str__(self):
        return f"{self.node_class} is not a Node"


Node_ = TypeVar("Node_", bound="Node")


@dataclass
class GlobalId(Generic[Node_]):
    id: str
    node_class: Type[Node_]

    @staticmethod
    def deserialize(global_id: ID) -> "GlobalId":
        try:
            node_class, id = global_id.split(":")
        except ValueError:
            raise InvalidGlobalId(global_id)
        if node_class not in _nodes:
            raise NotANode(node_class)
        return GlobalId(id, _nodes[node_class])

    def serialize(self) -> ID:
        return ID(f"{self.node_class._node_key()}:{self.id}")


deserializer(GlobalId.deserialize)
serializer(GlobalId.serialize)


# Use fake conversion to give the id field an ID type for Node (unspecialized) class
def _fake_serializer(_) -> GlobalId:
    raise NotImplementedError


N = TypeVar("N", bound="Node")


@interface
@dataclass  # type: ignore
class Node(Generic[Id], ABC):
    _id_type: ClassVar[AnyType]  # set in __init_subclass__
    id: Id = field(metadata=conversion(serialization=_fake_serializer))

    @classmethod
    def id_from_global(cls: Type[N], global_id: GlobalId[N]) -> Id:
        # Use coercion to handle integer id
        return cast(Id, deserialize(cls._id_type, global_id.id, coercion=True))

    @classmethod
    def id_to_global(cls: Type[N], id: Id) -> GlobalId[N]:
        return GlobalId(str(serialize(id)), cls)

    @property
    def global_id(self: N) -> GlobalId[N]:
        return self.id_to_global(self.id)

    @classmethod
    def get_by_global_id(
        cls: Type[N], global_id: GlobalId, info: graphql.GraphQLResolveInfo = None
    ) -> Union[N, Awaitable[N]]:
        if global_id.node_class != cls:
            raise ValueError(
                f"Expected {cls.__name__} id, found {global_id.node_class.__name__} id"
            )
        return cls.get_by_id(cls.id_from_global(global_id), info)

    @classmethod
    @abstractmethod
    def get_by_id(
        cls: Type[N], id: Id, info: graphql.GraphQLResolveInfo = None
    ) -> Union[N, Awaitable[N]]:
        raise NotImplementedError

    @classmethod
    def _node_key(cls) -> str:
        node_name = get_ref(cls)
        if node_name is None:
            raise TypeError(f"Node {cls} has no schema_ref registered")
        return node_name

    def __init_subclass__(cls, abstract: bool = False, **kwargs):
        super().__init_subclass__(**kwargs)  # type: ignore
        if abstract:
            return
        if sys.version_info >= (3, 7):
            if not has_type_vars(cls) and cls.get_by_id is not Node.get_by_id:
                _set_id_type(cls)
                _tmp_nodes.append(cls)
        else:
            if cls not in _tmp_nodes:
                _tmp_nodes.append(cls)


def _set_id_type(cls):
    for base in cls.__mro__:
        if base != Node and Node.get_by_id.__name__ in base.__dict__:
            if not isinstance(
                base.__dict__[base.get_by_id.__name__], (classmethod, staticmethod)
            ):
                raise TypeError(
                    f"{cls.__name__}.get_by_id must be a classmethod/staticmethod"
                )
            break
    else:
        raise TypeError(f"{cls.__name__}.{Node.get_by_id.__name__} must be defined")
    for base in generic_mro(cls):
        if get_origin(base) == Node:
            (_id_type,) = get_args(base)
            break
    else:
        raise NotImplementedError

    def serialize_id(id) -> GlobalId:
        return GlobalId(serialize(id), cls)

    id_field = cast(Field, getattr(cls, _FIELDS)["id"])
    cls.id = Field(  # type: ignore
        id_field.default,
        id_field.default_factory,  # type: ignore
        id_field.init,
        id_field.repr,
        id_field.hash,
        id_field.compare,
        id_field.metadata | conversion(serialization=serialize_id),
    )
    cls.__annotations__["id"] = _id_type
    cls._id_type = _id_type


# Use dict instead of set in order to keep order (because 3.6 generate duplicates)
_tmp_nodes: List[Type[Node]] = []
_nodes: Dict[str, Type[Node]] = {}


if sys.version_info < (3, 7):

    def nodes() -> Collection[Type[Node]]:
        for node_cls in _tmp_nodes:
            if (
                has_type_vars(node_cls)
                or get_args(node_cls)
                or node_cls.get_by_id is Node.get_by_id
            ):
                continue
            _set_id_type(node_cls)
            _nodes[node_cls._node_key()] = dataclass(node_cls)
        return list(_nodes.values())


else:

    def nodes() -> Collection[Type[Node]]:
        for node_cls in _tmp_nodes:
            _nodes[node_cls._node_key()] = node_cls
        return list(_nodes.values())


def node(id: GlobalId, info: graphql.GraphQLResolveInfo = None) -> Node:
    return id.node_class.get_by_global_id(id, info)



import collections.abc
import sys
import typing
from itertools import chain
from types import MappingProxyType
from typing import (
    AbstractSet,
    Any,
    Collection,
    Dict,
    FrozenSet,
    List,
    Mapping,
    MutableMapping,
    MutableSequence,
    MutableSet,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
)

from apischema.typing import get_origin

AnyType = Any
NoneType: Type[None] = type(None)
Number = Union[int, float]

PRIMITIVE_TYPES = (str, int, bool, float, NoneType)
COLLECTION_TYPES = {
    Collection: tuple,
    collections.abc.Collection: tuple,
    Sequence: tuple,
    collections.abc.Sequence: tuple,
    Tuple: tuple,
    tuple: tuple,
    MutableSequence: list,
    collections.abc.MutableSequence: list,
    List: list,
    list: list,
    AbstractSet: frozenset,
    collections.abc.Set: frozenset,
    FrozenSet: frozenset,
    frozenset: frozenset,
    MutableSet: set,
    collections.abc.MutableSet: set,
    Set: set,
    set: set,
}
MAPPING_TYPES = {
    Mapping: MappingProxyType,
    collections.abc.Mapping: MappingProxyType,
    MutableMapping: dict,
    collections.abc.MutableMapping: dict,
    Dict: dict,
    dict: dict,
    MappingProxyType: MappingProxyType,
}


if (3, 7) <= sys.version_info < (3, 9):  # pragma: no cover

    def subscriptable_origin(tp: AnyType) -> AnyType:
        if (
            type(tp) == type(List[int])  # noqa: E721
            and tp.__module__ == "typing"
            and hasattr(tp, "_name")
        ):
            return getattr(typing, tp._name)
        else:
            return get_origin(tp)


else:  # pragma: no cover
    subscriptable_origin = get_origin  # type: ignore


if sys.version_info >= (3, 7):  # pragma: no cover
    OrderedDict = dict
    ChainMap = collections.ChainMap
else:  # pragma: no cover
    OrderedDict = collections.OrderedDict

    class ChainMap(collections.ChainMap):
        def __iter__(self):
            return iter({k: None for k in chain.from_iterable(reversed(self.maps))})


class MetadataUnion(Mapping[str, Any]):
    def __or__(self, other: Mapping[str, Any]) -> "Metadata":
        return MappingWithUnion({**self, **other})

    def __ror__(self, other: Mapping[str, Any]) -> "Metadata":
        return MappingWithUnion({**other, **self})


# Kind of hack to benefit of PEP 584
if sys.version_info >= (3, 9):  # pragma: no cover
    Metadata = Mapping[str, Any]
else:  # pragma: no cover
    Metadata = MetadataUnion


class MetadataMixin(MetadataUnion):
    key: str

    def __getitem__(self, key):
        if key != self.key:
            raise KeyError(key)
        return self

    def __iter__(self):
        return iter((self.key,))

    def __len__(self):
        return 1


if sys.version_info >= (3, 9):  # pragma: no cover
    MappingWithUnion = MappingProxyType
else:  # pragma: no cover

    class MappingWithUnion(dict, MetadataUnion):
        pass



__all__ = [
    "ClientMutationId",
    "Connection",
    "Edge",
    "GlobalId",
    "Mutation",
    "Node",
    "PageInfo",
    "base64_encoding",
    "mutations",
    "node",
    "nodes",
]
from .connections import Connection, Edge, PageInfo
from .global_identification import GlobalId, Node, node, nodes
from .mutations import ClientMutationId, Mutation, mutations
from .utils import base64_encoding


def alter_default_ref():
    from apischema import settings
    from apischema.graphql.relay.connections import Connection, Edge, PageInfo
    from apischema.graphql.relay.global_identification import Node
    from apischema.json_schema.refs import get_ref
    from apischema.types import NoneType
    from apischema.typing import generic_mro, get_args, get_origin
    from apischema.utils import (
        get_args2,
        get_origin_or_type,
        has_type_vars,
        is_union_of,
    )

    _set_ref = settings.default_ref
    _origin_default_ref = settings.default_ref()

    def get_node_ref(tp):
        if is_union_of(tp, NoneType) and len(get_args2(tp)):
            tp = next(arg for arg in get_args2(tp) if arg is not NoneType)
        ref = get_ref(tp)
        if ref is None:
            raise TypeError(
                f"Node {tp} must have a ref registered to be used with connection"
            )
        return ref

    def _default_ref(func=None):
        nonlocal _origin_default_ref
        if func is None:
            return _origin_default_ref
        else:
            _origin_default_ref = func

            def wrapper(tp):
                cls = get_origin_or_type(tp)
                if (
                    isinstance(cls, type)
                    and issubclass(cls, (Connection, Edge, PageInfo))
                    and not has_type_vars(tp)
                ):
                    for base in generic_mro(tp):
                        if get_origin(base) == Connection:
                            return f"{get_node_ref(get_args(base)[0])}Connection"
                        if get_origin(base) == Edge:
                            return f"{get_node_ref(get_args(base)[0])}Edge"
                        if get_origin(base) == PageInfo:
                            return "PageInfo"
                elif cls == Node:
                    return "Node"
                else:
                    return func(tp)

            _set_ref(wrapper)

    settings.default_ref = _default_ref
    _default_ref(_origin_default_ref)


alter_default_ref()
del alter_default_ref



from dataclasses import dataclass, fields
from typing import Any, Dict, Mapping, Optional, Sequence, Union

from apischema.utils import Undefined, merge_opts, merge_opts_mapping

Deprecated = Union[bool, str]


@dataclass(frozen=True)
class Annotations:
    extra: Optional[Mapping[str, Any]] = None
    title: Optional[str] = None
    description: Optional[str] = None
    default: Optional[Any] = Undefined
    examples: Optional[Sequence[Any]] = None
    format: Optional[str] = None
    deprecated: Deprecated = False

    def as_dict(self) -> Mapping[str, Any]:
        result: Dict[str, Any] = {}
        if self.deprecated:
            result["deprecated"] = self.deprecated
        for k in ("title", "description", "examples", "format"):
            if getattr(self, k) is not None:
                result[k] = getattr(self, k)
        if self.default is not Undefined:
            result["default"] = self.default
        if self.extra:
            result.update(self.extra)
        return result


@merge_opts
def merge_annotations(default: Annotations, override: Annotations) -> Annotations:
    merged_fields = {
        f.name: getattr(override, f.name) or getattr(default, f.name)
        for f in fields(Annotations)
    }
    if override.default is not Undefined:
        merged_fields["default"] = override.default
    else:
        merged_fields["default"] = default.default
    merged_fields["extra"] = merge_opts_mapping(default.extra, override.extra)  # type: ignore # noqa: E501
    return Annotations(**merged_fields)



from collections import defaultdict
from dataclasses import dataclass
from functools import wraps
from inspect import Parameter, isclass, signature
from typing import (
    Any,
    Callable,
    Dict,
    Mapping,
    NoReturn,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

from apischema.conversions.conversions import Conversions
from apischema.conversions.dataclass_models import get_model_origin, has_model_origin
from apischema.json_schema.schema import Schema
from apischema.types import AnyType
from apischema.typing import generic_mro, get_args, get_type_hints
from apischema.utils import (
    MethodOrProperty,
    MethodWrapper,
    Undefined,
    UndefinedType,
    get_args2,
    get_origin2,
    get_origin_or_type,
    get_parameters,
    is_method,
    method_class,
    method_wrapper,
    substitute_type_vars,
)


@dataclass(frozen=True)
class Serialized:
    func: Callable
    conversions: Optional[Conversions]
    schema: Optional[Schema]
    error_handler: Optional[Callable]

    def error_type(self) -> AnyType:
        assert self.error_handler is not None
        types = get_type_hints(self.error_handler, include_extras=True)
        if "return" not in types:
            raise TypeError("Error handler must be typed")
        return types["return"]

    def return_type(self, return_type: AnyType) -> AnyType:
        if self.error_handler is not None:
            error_type = self.error_type()
            if error_type is not NoReturn:
                return Union[return_type, error_type]
        return return_type

    def types(self, owner: AnyType = None) -> Mapping[str, AnyType]:
        types = get_type_hints(self.func, include_extras=True)
        if "return" not in types:
            if isclass(self.func):
                types["return"] = self.func
            else:
                raise TypeError("Function must be typed")
        types["return"] = self.return_type(types["return"])
        if get_args2(owner):
            substitution = dict(
                zip(get_parameters(get_origin2(owner)), get_args2(owner))
            )
            types = {
                name: substitute_type_vars(tp, substitution)
                for name, tp in types.items()
            }
        return types


_serialized_methods: Dict[Type, Dict[str, Serialized]] = defaultdict(dict)

S = TypeVar("S", bound=Serialized)


def _get_methods(
    tp: AnyType, all_methods: Mapping[Type, Mapping[str, S]]
) -> Mapping[str, Tuple[S, Mapping[str, AnyType]]]:
    result = {}
    for base in reversed(generic_mro(tp)):
        for name, method in all_methods[get_origin_or_type(base)].items():
            result[name] = (method, method.types(base))
    if has_model_origin(tp):
        origin = get_model_origin(tp)
        if get_args2(tp):
            substitution = dict(zip(get_parameters(tp), get_args(tp)))
            origin = substitute_type_vars(origin, substitution)
        result.update(_get_methods(origin, all_methods))
    return result


def get_serialized_methods(
    tp: AnyType,
) -> Mapping[str, Tuple[Serialized, Mapping[str, AnyType]]]:
    return _get_methods(tp, _serialized_methods)


ErrorHandler = Union[Callable, None, UndefinedType]


def none_error_handler(error: Exception, obj: Any, alias: str) -> None:
    return None


def register_serialized(
    func: Callable,
    alias: str,
    conversions: Optional[Conversions],
    schema: Optional[Schema],
    error_handler: ErrorHandler,
    owner: Type = None,
):
    parameters = list(signature(func).parameters.values())
    for param in parameters[1:]:
        if (
            param.kind not in {Parameter.VAR_POSITIONAL, Parameter.VAR_KEYWORD}
            and param.default is Parameter.empty
        ):
            raise TypeError("Serialized method cannot have parameter without default")
    if error_handler is None:
        error_handler = none_error_handler
    if error_handler is Undefined:
        error_handler = None
    else:
        wrapped = func

        @wraps(wrapped)
        def func(self):
            try:
                return wrapped(self)
            except Exception as error:
                return error_handler(error, self, alias)

    assert not isinstance(error_handler, UndefinedType)
    serialized = Serialized(func, conversions, schema, error_handler)
    if owner is None:
        try:
            owner = get_origin_or_type(get_type_hints(func)[parameters[0].name])
        except KeyError:
            raise TypeError(
                "First parameter of serialized method must be typed"
            ) from None
    _serialized_methods[owner][alias] = serialized


class SerializedDescriptor(MethodWrapper[MethodOrProperty]):
    def __init__(
        self,
        method: MethodOrProperty,
        alias: Optional[str],
        conversions: Optional[Conversions],
        schema: Optional[Schema],
        error_handler: ErrorHandler,
    ):
        super().__init__(method)
        self._alias = alias
        self._conversions = conversions
        self._schema = schema
        self._error_handler = error_handler

    def __set_name__(self, owner, name):
        super().__set_name__(owner, name)
        register_serialized(
            method_wrapper(self._method, name),
            self._alias or name,
            self._conversions,
            self._schema,
            self._error_handler,
            owner,
        )

    def __call__(self, *args, **kwargs):
        raise RuntimeError("Method __set_name__ has not been called")


MethodOrProp = TypeVar("MethodOrProp", Callable, property)


@overload
def serialized(__method_or_property: MethodOrProp) -> MethodOrProp:
    ...


@overload
def serialized(
    alias: str = None,
    *,
    conversions: Conversions = None,
    schema: Schema = None,
    error_handler: ErrorHandler = Undefined,
    owner: Type = None,
) -> Callable[[MethodOrProp], MethodOrProp]:
    ...


def serialized(
    __arg=None,
    *,
    alias: str = None,
    conversions: Conversions = None,
    schema: Schema = None,
    error_handler: ErrorHandler = Undefined,
    owner: Type = None,
):
    def decorator(method: MethodOrProperty):
        nonlocal owner
        if is_method(method) and method_class(method) is None:
            return SerializedDescriptor(
                method, alias, conversions, schema, error_handler
            )
        else:
            if is_method(method):
                if owner is None:
                    owner = method_class(method)
                method = method_wrapper(method)
            assert not isinstance(method, property)
            register_serialized(
                cast(Callable, method),
                alias or method.__name__,
                conversions,
                schema,
                error_handler,
                owner,
            )
            return method

    if isinstance(__arg, str) or __arg is None:
        alias = alias or __arg
        return decorator
    else:
        return decorator(__arg)



import re
from dataclasses import dataclass
from typing import Callable, Optional, Pattern, TYPE_CHECKING, Tuple, Union

from apischema.metadata.keys import (
    CONVERSIONS_METADATA,
    DEFAULT_AS_SET,
    DEFAULT_FALLBACK_METADATA,
    INIT_VAR_METADATA,
    MERGED_METADATA,
    POST_INIT_METADATA,
    PROPERTIES_METADATA,
    REQUIRED_METADATA,
    SKIP_METADATA,
    VALIDATORS_METADATA,
)
from apischema.types import AnyType, MappingWithUnion, Metadata, MetadataMixin

if TYPE_CHECKING:
    from apischema.conversions.conversions import ConvOrFunc
    from apischema.validation.validator import Validator


def simple_metadata(key: str) -> Metadata:
    return MappingWithUnion({key: ...})


@dataclass
class ConversionMetadata(MetadataMixin):
    key = CONVERSIONS_METADATA
    deserialization: Optional["ConvOrFunc"] = None
    serialization: Optional["ConvOrFunc"] = None


if False:  # For Pycharm

    def conversion(
        deserialization: "ConvOrFunc" = None,
        serialization: "ConvOrFunc" = None,
    ) -> ConversionMetadata:
        ...


else:
    conversion = ConversionMetadata


default_as_set = simple_metadata(DEFAULT_AS_SET)

default_fallback = simple_metadata(DEFAULT_FALLBACK_METADATA)


def init_var(tp: AnyType) -> Metadata:
    return MappingWithUnion({INIT_VAR_METADATA: tp})


merged = simple_metadata(MERGED_METADATA)

post_init = simple_metadata(POST_INIT_METADATA)


class PropertiesMetadata(dict, Metadata):
    def __init__(self):
        super().__init__({PROPERTIES_METADATA: None})

    def __call__(
        self, pattern: Union[str, Pattern, "ellipsis"]  # noqa: F821
    ) -> Metadata:
        if pattern is not ...:
            pattern = re.compile(pattern)
        return MappingWithUnion({PROPERTIES_METADATA: pattern})


properties = PropertiesMetadata()


required = simple_metadata(REQUIRED_METADATA)

skip = simple_metadata(SKIP_METADATA)


@dataclass(frozen=True)
class ValidatorsMetadata(MetadataMixin):
    key = VALIDATORS_METADATA
    validators: Tuple["Validator", ...]


def validators(*validator: Callable) -> ValidatorsMetadata:
    from apischema.validation.validator import Validator

    return ValidatorsMetadata(tuple(map(Validator, validator)))



from typing import Pattern

from apischema.types import AnyType


def infer_pattern(tp: AnyType) -> Pattern:
    from apischema.json_schema.generation.schema import DeserializationSchemaBuilder

    try:
        builder = DeserializationSchemaBuilder(lambda s: s, lambda s: s, {}, False)
        prop_schema = builder.visit(tp)
    except RecursionError:
        pass
    else:
        if (
            len(prop_schema.get("patternProperties", {})) == 1
            and "additionalProperties" not in prop_schema
        ):
            return next(iter(prop_schema["patternProperties"]))
    raise TypeError("Cannot inferred pattern from type schema") from None



from contextlib import contextmanager
from types import new_class
from typing import (
    Any,
    Collection,
    Generic,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from apischema.conversions.conversions import (
    Conversions,
    HashableConversions,
    ResolvedConversion,
    resolve_conversions,
    resolve_deserialization,
    resolve_serialization,
    to_hashable_conversions,
)
from apischema.conversions.converters import _deserializers, _serializers
from apischema.conversions.dataclass_models import DataclassModel
from apischema.conversions.utils import is_convertible
from apischema.dataclasses import replace
from apischema.skip import filter_skipped
from apischema.types import AnyType
from apischema.typing import get_args, get_origin
from apischema.utils import (
    OperationKind,
    Undefined,
    UndefinedType,
    get_parameters,
    substitute_type_vars,
)
from apischema.visitor import Return, Visitor

try:
    from apischema.typing import Annotated
except ImportError:
    Annotated = ...  # type: ignore

Deserialization = Sequence[ResolvedConversion]
Serialization = ResolvedConversion
Conv = TypeVar("Conv", Deserialization, Serialization)


class ConversionsVisitor(Visitor[Return], Generic[Conv, Return]):
    operation: OperationKind
    _dynamic_conversion_resolver: Type["DynamicConversionResolver"]

    def __init__(self):
        super().__init__()
        self._conversions: Optional[HashableConversions] = None

    @staticmethod
    def _get_conversions(
        tp: Type, conversions: Conversions
    ) -> Union[Conv, None, UndefinedType]:
        raise NotImplementedError

    @staticmethod
    def _default_conversions(tp: Type) -> Optional[Conversions]:
        raise NotImplementedError

    @classmethod
    def get_conversions(
        cls, tp: Type, conversions: Optional[Conversions]
    ) -> Optional[Conv]:
        result = None
        if conversions is not None:
            result = cls._get_conversions(tp, conversions)
        if result is None:
            conversions = cls._default_conversions(tp)
            if conversions is not None:
                result = cls._get_conversions(tp, conversions)
        return result if result is not Undefined else None  # type: ignore

    def _apply_dynamic_conversions(self, tp: AnyType) -> Optional[AnyType]:
        if self._conversions is None:
            return None
        else:
            return self._dynamic_conversion_resolver().visit_with_conversions(
                tp, self._conversions
            )

    def visit_conversion(self, cls: Type, conversion: Conv) -> Return:
        raise NotImplementedError

    def visit_not_conversion(self, tp: AnyType) -> Return:
        return super()._visit(tp)

    @contextmanager
    def _replace_conversions(self, conversions: Optional[Conversions]):
        conversions_save = self._conversions
        self._conversions = to_hashable_conversions(conversions)
        try:
            yield
        finally:
            self._conversions = conversions_save

    def _visit(self, tp: AnyType) -> Return:
        if not is_convertible(tp):
            if isinstance(tp, DataclassModel):
                return self.visit(tp.dataclass)
            else:
                return self.visit_not_conversion(tp)
        conversion = self.get_conversions(tp, self._conversions)
        with self._replace_conversions(None):
            if conversion is not None:
                return self.visit_conversion(tp, conversion)
            else:
                return self.visit_not_conversion(tp)

    def _replace_generic_args(self, tp: AnyType) -> AnyType:
        if self._generic is not None:
            substitution = dict(
                zip(get_parameters(get_origin(self._generic)), get_args(self._generic))
            )
            return substitute_type_vars(tp, substitution)
        else:
            return tp

    def visit_with_conversions(
        self, tp: AnyType, conversions: Optional[Conversions]
    ) -> Return:
        with self._replace_conversions(conversions):
            return self.visit(tp)


class DeserializationVisitor(ConversionsVisitor[Deserialization, Return]):
    operation = OperationKind.DESERIALIZATION

    @staticmethod
    def _get_conversions(
        tp: Type, conversions: Conversions
    ) -> Union[Deserialization, None, UndefinedType]:
        resolved_conversions = [
            conv
            for conv in resolve_conversions(conversions, resolve_deserialization)
            if conv.target == tp
        ]
        for i, conv in enumerate(resolved_conversions):
            if conv.is_identity and conv.source == tp:
                if len(resolved_conversions) == 1:
                    return Undefined
                else:
                    namespace = {
                        "__new__": lambda _, *args, **kwargs: tp(*args, **kwargs)
                    }
                    wrapper = new_class(
                        tp.__name__, (tp,), exec_body=lambda ns: ns.update(namespace)
                    )
                    resolved_conversions[i] = replace(conv, source=wrapper)
        return resolved_conversions or None

    _default_conversions = staticmethod(_deserializers.get)  # type: ignore

    def visit_conversion(self, cls: Type, conversion: Deserialization) -> Return:
        return self._union_result(
            [
                self.visit_with_conversions(
                    self._replace_generic_args(conv.source), conv.conversions
                )
                for conv in conversion
            ]
        )


class SerializationVisitor(ConversionsVisitor[Serialization, Return]):
    operation = OperationKind.SERIALIZATION

    @staticmethod
    def _get_conversions(
        tp: Type, conversions: Conversions
    ) -> Union[Serialization, None, UndefinedType]:
        for conv in resolve_conversions(conversions, resolve_serialization):
            if issubclass(tp, conv.source):
                if conv.is_identity and conv.target == tp:
                    return Undefined
                return conv
        else:
            return None

    @staticmethod
    def _default_conversions(tp: Type) -> Optional[Conversions]:
        for sub_cls in tp.__mro__:
            if sub_cls in _serializers:
                return _serializers[sub_cls]
        else:
            return None

    def visit_conversion(self, cls: Type, conversion: Serialization) -> Return:
        return self.visit_with_conversions(
            self._replace_generic_args(conversion.target), conversion.conversions
        )


class DynamicConversionResolver(ConversionsVisitor[Conv, Optional[AnyType]]):
    def __init__(self):
        super().__init__()
        self._rec_guard: Set[HashableConversions] = set()

    def annotated(self, tp: AnyType, annotations: Sequence[Any]) -> Optional[AnyType]:
        result = self.visit(tp)
        return Annotated[(result, *annotations)]

    def collection(
        self, cls: Type[Collection], value_type: AnyType
    ) -> Optional[AnyType]:
        value = self.visit(value_type)
        return None if value is None else Sequence[value]  # type: ignore

    def mapping(
        self, cls: Type[Mapping], key_type: AnyType, value_type: AnyType
    ) -> Optional[AnyType]:
        key = self.visit(key_type)
        value = self.visit(value_type)
        return None if key is None or value is None else Mapping[key, value]  # type: ignore # noqa: E501

    def visit_types(
        self, types: Iterable[AnyType], origin: AnyType
    ) -> Optional[AnyType]:
        modified = False
        types2 = []
        for tp in types:
            try:
                res = self.visit(tp)
                types2.append(res if res is not None else tp)
                modified = True
            except Exception:
                types2.append(tp)
        return origin[types] if modified else None

    def tuple(self, types: Sequence[AnyType]) -> Optional[AnyType]:
        return self.visit_types(types, Tuple)

    def union(self, alternatives: Sequence[AnyType]) -> Optional[AnyType]:
        return self.visit_types(filter_skipped(alternatives, schema_only=True), Union)

    def _final_type(self, conversion: Conv) -> Optional[AnyType]:
        raise NotImplementedError

    def _visit(self, tp: AnyType) -> Optional[AnyType]:
        if (
            self._conversions is None
            or not is_convertible(tp)
            or self._conversions in self._rec_guard
        ):
            return None
        conv = self._get_conversions(tp, self._conversions)
        if conv is None or conv is Undefined:
            return None
        self._rec_guard.add(self._conversions)
        try:
            result = self.visit_conversion(tp, conv)
            return result if result is not None else self._final_type(conv)
        except Exception:
            return self._final_type(conv)
        finally:
            self._rec_guard.remove(self._conversions)

    def visit(self, tp: AnyType) -> Optional[AnyType]:
        try:
            return super().visit(tp)
        except Exception:
            return None


class DynamicDeserializationResolver(DynamicConversionResolver, DeserializationVisitor):
    def _final_type(self, conversion: Deserialization) -> Optional[AnyType]:
        return Union[
            tuple(self._replace_generic_args(conv.source) for conv in conversion)
        ]


class DynamicSerializationResolver(DynamicConversionResolver, SerializationVisitor):
    def _final_type(self, conversion: Serialization) -> Optional[AnyType]:
        return self._replace_generic_args(conversion.target)


DeserializationVisitor._dynamic_conversion_resolver = DynamicDeserializationResolver
SerializationVisitor._dynamic_conversion_resolver = DynamicSerializationResolver



from inspect import Parameter, isclass, signature
from typing import Any, Callable, Dict, Generic, Optional, Tuple, Type, TypeVar, cast

from apischema.types import AnyType, COLLECTION_TYPES, MAPPING_TYPES, PRIMITIVE_TYPES
from apischema.typing import get_args, get_origin, get_type_hints
from apischema.utils import get_parameters, is_type_var, substitute_type_vars, type_name

try:
    from apischema.typing import Annotated
except ImportError:
    Annotated = ...  # type: ignore

Converter = Callable[[Any], Any]


def converter_types(
    converter: Converter,
    source: Optional[AnyType] = None,
    target: Optional[AnyType] = None,
    namespace: Dict[str, Any] = None,
) -> Tuple[AnyType, AnyType]:
    try:
        # in pre 3.9, Generic __new__ perturb signature of types
        if (
            isinstance(converter, type)
            and converter.__new__ is Generic.__new__ is not object.__new__
            and converter.__init__ is not object.__init__  # type: ignore
        ):
            parameters = list(signature(converter.__init__).parameters.values())[1:]  # type: ignore # noqa: E501
        else:
            parameters = list(signature(converter).parameters.values())
    except ValueError:  # builtin types
        if target is None and isclass(converter):
            target = cast(Type[Any], converter)
        if source is None:
            raise TypeError("Converter source is unknown") from None
    else:
        if not parameters:
            raise TypeError("converter must have at least one parameter")
        first_param, *other_params = parameters
        for p in other_params:
            if p.default is Parameter.empty and p.kind not in (
                Parameter.VAR_POSITIONAL,
                Parameter.VAR_KEYWORD,
            ):
                raise TypeError(
                    "converter must have at most one parameter without default"
                )
        if source is not None and target is not None:
            return source, target
        types = get_type_hints(converter, None, namespace, include_extras=True)
        if not types and isclass(converter):
            types = get_type_hints(
                converter.__new__, None, namespace, include_extras=True
            ) or get_type_hints(
                converter.__init__, None, namespace, include_extras=True  # type: ignore
            )
        if source is None:
            try:
                source = types.pop(first_param.name)
            except KeyError:
                raise TypeError("converter source is unknown") from None
        if target is None:
            try:
                target = types.pop("return")
            except KeyError:
                if isclass(converter):
                    target = cast(Type, converter)
                else:
                    raise TypeError("converter target is unknown") from None
    return source, target


def get_conversion_type(base: AnyType, other: AnyType) -> Tuple[AnyType, AnyType]:
    """
    Args:
        base: (generic) type on the registered side the conversion
        other: other side of the conversion

    Returns:
        The type on which is registered conversion (its origin when generic) and the
        other side of the conversion with its original parameters
    """
    origin = get_origin(base)
    if origin is Annotated:
        raise TypeError("Annotated types cannot have conversions")
    if origin is None:
        return base, other
    args = get_args(base)
    if not all(map(is_type_var, args)):
        raise TypeError(
            f"Generic conversion doesn't support specialization,"
            f" aka {type_name(base)}[{','.join(map(type_name, args))}]"
        )
    return origin, substitute_type_vars(other, dict(zip(args, get_parameters(origin))))


BUILTIN_TYPES = {*PRIMITIVE_TYPES, *COLLECTION_TYPES, *MAPPING_TYPES}


def is_convertible(tp: AnyType) -> bool:
    return isinstance(tp, type) and tp not in BUILTIN_TYPES


T = TypeVar("T")


def identity(x: T) -> T:
    return x



__all__ = [
    "Discard",
    "ValidationError",
    "ValidatorResult",
    "get_validators",
    "validate",
    "validator",
    "with_validation_error",
]

from .errors import ValidationError, ValidatorResult, with_validation_error
from .validator import Discard, get_validators, validate, validator



from dataclasses import Field
from typing import Mapping

from apischema.utils import PREFIX

ALIAS_METADATA = f"{PREFIX}alias"
ALIAS_NO_OVERRIDE_METADATA = f"{PREFIX}alias_no_override"
CONVERSIONS_METADATA = f"{PREFIX}conversions"
DEFAULT_AS_SET = f"{PREFIX}default_as_set"
DEFAULT_FALLBACK_METADATA = f"{PREFIX}default_fallback"
INIT_VAR_METADATA = f"{PREFIX}init_var"
MERGED_METADATA = f"{PREFIX}merged"
POST_INIT_METADATA = f"{PREFIX}post_init"
PROPERTIES_METADATA = f"{PREFIX}properties"
REQUIRED_METADATA = f"{PREFIX}required"
SCHEMA_METADATA = f"{PREFIX}schema"
SKIP_METADATA = f"{PREFIX}skip"
VALIDATORS_METADATA = f"{PREFIX}validators"

FORBIDDEN_WITH_AGGREGATE = {
    ALIAS_METADATA,
    ALIAS_NO_OVERRIDE_METADATA,
    DEFAULT_AS_SET,
    POST_INIT_METADATA,
    REQUIRED_METADATA,
    SCHEMA_METADATA,
    VALIDATORS_METADATA,
}


def is_aggregate_field(field: Field) -> bool:
    return MERGED_METADATA in field.metadata or PROPERTIES_METADATA in field.metadata


def check_metadata(field: Field) -> Mapping:
    if MERGED_METADATA in field.metadata and PROPERTIES_METADATA in field.metadata:
        raise TypeError("merged and properties metadata are incompatible")
    if is_aggregate_field(field):
        forbidden = FORBIDDEN_WITH_AGGREGATE & field.metadata.keys()
        if forbidden:
            raise TypeError(f"{forbidden} metadata are not allowed in aggregate field")
    return field.metadata



__all__ = [
    "additional_properties",
    "aliaser",
    "coercer",
    "coercion",
    "default_fallback",
    "deserialization",
    "json_schema_version",
    "serialization",
]

from typing import Callable, Optional, Type, TypeVar, overload

from apischema import aliases
from apischema.aliases import Aliaser
from apischema.cache import reset_cache
from apischema.conversions.conversions import Conversions
from apischema.conversions.visitor import (
    DeserializationVisitor,
    SerializationVisitor,
)
from apischema.deserialization import coercion as coercion_
from apischema.json_schema import refs, schema, versions
from apischema.types import AnyType
from apischema.utils import to_camel_case

additional_properties = False
coercion: bool = False
default_fallback = False
exclude_unset = True
json_schema_version = versions.JsonSchemaVersion.DRAFT_2019_09


@overload
def coercer() -> "coercion_.Coercer":
    ...


@overload
def coercer(func: "coercion_.Coercer") -> "coercion_.Coercer":
    ...


def coercer(func=None):
    if func is None:
        return coercion_._coercer
    elif func is coercion_.coerce:
        coercion_._coercer = func
    else:
        coercion_._coercer = coercion_.wrap_coercer(func)
        return func


ConversionsFunc = Callable[[Type], Optional[Conversions]]


@overload
def deserialization() -> ConversionsFunc:
    ...


@overload
def deserialization(func: ConversionsFunc) -> ConversionsFunc:
    ...


def deserialization(func=None):
    if func is None:
        return DeserializationVisitor._default_conversions
    else:
        DeserializationVisitor._default_conversions = staticmethod(func)
        reset_cache()
        return func


@overload
def serialization() -> ConversionsFunc:
    ...


@overload
def serialization(func: ConversionsFunc) -> ConversionsFunc:
    ...


def serialization(func=None):
    if func is None:
        return SerializationVisitor._default_conversions
    else:
        SerializationVisitor._default_conversions = staticmethod(func)
        reset_cache()
        return func


AliaserFunc = TypeVar("AliaserFunc", bound=Aliaser)


@overload
def aliaser() -> Aliaser:
    ...


@overload
def aliaser(func: AliaserFunc) -> AliaserFunc:
    ...


@overload
def aliaser(*, camel_case: bool):
    ...


def aliaser(func=None, *, camel_case: bool = None):
    if camel_case is True:
        func = to_camel_case
    elif camel_case is False:
        func = lambda s: s  # noqa: E731
    if func is None:
        return aliases._global_aliaser
    else:
        aliases._global_aliaser = func
    return func


RefFunc = Callable[[AnyType], refs.Ref]


@overload
def default_ref() -> RefFunc:
    ...


@overload
def default_ref(func: RefFunc) -> RefFunc:
    ...


def default_ref(func=None):
    if func is None:
        return refs._default_ref
    else:
        refs._default_ref = func


SchemaFunc = Callable[[AnyType], Optional[schema.Schema]]


@overload
def default_schema() -> SchemaFunc:
    ...


@overload
def default_schema(func: SchemaFunc) -> SchemaFunc:
    ...


def default_schema(func=None):
    if func is None:
        return schema._default_schema
    else:
        schema._default_schema = func
        reset_cache()



from enum import Enum, auto

from apischema.conversions.conversions import (
    ConvOrFunc,
    ResolvedConversion,
    handle_serialization_method,
    to_conversion,
)
from apischema.conversions.utils import converter_types
from apischema.dataclasses import replace
from apischema.types import AnyType
from apischema.typing import get_args, get_origin
from apischema.utils import get_args2, get_origin2, is_type_var, substitute_type_vars

try:
    from apischema.typing import Annotated
except ImportError:
    Annotated = ...  # type: ignore


class Variance(Enum):
    COVARIANT = auto()
    CONTRAVARIANT = auto()


def handle_generic_field_type(
    field_type: AnyType, base: AnyType, other: AnyType, variance: Variance
) -> AnyType:
    """When conversion is generic, try to adapt the other side of the conversion
    depending on the field_type

    Args:
        field_type: type of the field
        base: side of the conversion where the field is
        other: other side of the conversion
        variance: variance of the conversion
    """
    if is_type_var(base):
        type_vars = {base: field_type}
    elif (
        # field_type is generic with free typevars, field_type also generic with
        # number of args, none of their args are generic (must be type or typevar)
        # and with
        get_origin2(base) is not None
        and getattr(base, "__parameters__", ())
        and len(get_args2(base)) == len(get_args2(field_type))
        and not any(map(get_origin2, get_args2(base)))
        and not any(map(get_origin2, get_args2(field_type)))
        and not any(
            not is_type_var(base_arg) and base_arg != field_arg
            for base_arg, field_arg in zip(get_args2(base), get_args2(field_type))
        )
    ):
        field_type_origin, base_origin = get_origin2(field_type), get_origin2(base)
        assert field_type_origin is not None and base_origin is not None
        if base_origin != field_type_origin:
            if variance == Variance.COVARIANT and not issubclass(
                base_origin, field_type_origin
            ):
                return other
            if variance == Variance.CONTRAVARIANT and not issubclass(
                field_type_origin, base_origin
            ):
                return other
        type_vars = {}
        for base_arg, field_arg in zip(get_args2(base), get_args2(field_type)):
            if base_arg in type_vars and type_vars[base_arg] != field_arg:
                return other
            type_vars[base_arg] = field_arg
    else:
        return other
    if get_origin(base) is Annotated:
        other = Annotated[(other, *get_args(base)[1:])]
    return substitute_type_vars(other, type_vars)


def resolve_field_deserialization(
    field_type: AnyType, conversion: ConvOrFunc
) -> ResolvedConversion:
    conversion = to_conversion(conversion)
    if isinstance(conversion.converter, property):
        raise TypeError("Field deserialization cannot be a property")
    try:
        source, target = converter_types(
            conversion.converter, conversion.source, conversion.target
        )
    except TypeError:
        if conversion.target is None:
            source, target = converter_types(
                conversion.converter, conversion.source, field_type
            )
        else:
            raise
    else:
        source = handle_generic_field_type(
            field_type, target, source, Variance.COVARIANT
        )
    return ResolvedConversion.from_conversion(
        replace(conversion, source=source, target=target)
    )


def resolve_field_serialization(
    field_type: AnyType, conversion: ConvOrFunc
) -> ResolvedConversion:
    conversion = handle_serialization_method(conversion)
    assert not isinstance(conversion.converter, property)
    try:
        source, target = converter_types(
            conversion.converter, conversion.source, conversion.target
        )
    except TypeError:
        if conversion.source is None:
            source, target = converter_types(
                conversion.converter, field_type, conversion.target
            )
        else:
            raise
    else:
        target = handle_generic_field_type(
            field_type, source, target, Variance.CONTRAVARIANT
        )
    return ResolvedConversion.from_conversion(
        replace(conversion, source=source, target=target)
    )



from dataclasses import Field
from typing import Any, Iterator, Mapping, Sequence, Tuple, Type

from apischema.conversions.visitor import Deserialization, DeserializationVisitor
from apischema.dataclass_utils import get_alias, get_field_conversion, get_fields
from apischema.metadata.keys import MERGED_METADATA
from apischema.types import AnyType
from apischema.utils import OperationKind
from apischema.visitor import Unsupported


class InitMergedAliasVisitor(DeserializationVisitor[Iterator[str]]):
    def dataclass(
        self,
        cls: Type,
        types: Mapping[str, AnyType],
        fields: Sequence[Field],
        init_vars: Sequence[Field],
    ) -> Iterator[str]:
        for field in get_fields(
            fields, init_vars, operation=OperationKind.DESERIALIZATION
        ):
            if MERGED_METADATA in field.metadata:
                yield from get_init_merged_alias(cls, field, types[field.name])
            else:
                yield get_alias(field)

    def mapping(
        self, cls: Type[Mapping], key_type: AnyType, value_type: AnyType
    ) -> Iterator[str]:
        yield from ()

    def named_tuple(
        self,
        cls: Type[Tuple],
        types: Mapping[str, AnyType],
        defaults: Mapping[str, Any],
    ) -> Iterator[str]:
        yield from types

    def typed_dict(
        self, cls: Type, keys: Mapping[str, AnyType], total: bool
    ) -> Iterator[str]:
        yield from keys

    def visit_conversion(self, cls: Type, conversion: Deserialization) -> Iterator[str]:
        if len(conversion) != 1:
            raise NotImplementedError
        conv = conversion[0]
        return self.visit_with_conversions(conv.source, conv.conversions)


def get_init_merged_alias(
    cls: Type, field: Field, field_type: AnyType
) -> Iterator[str]:
    field_type, conversion = get_field_conversion(
        field, field_type, OperationKind.DESERIALIZATION
    )
    try:
        yield from InitMergedAliasVisitor().visit_with_conversions(
            field_type, conversion.conversions if conversion is not None else None
        )
    except (NotImplementedError, Unsupported):
        raise TypeError(
            f"Merged field {cls.__name__}.{field.name} must have an object type"
            f" or an unique deserializer to an object type"
        ) from None



from dataclasses import dataclass
from typing import (
    Callable,
    Collection,
    Optional,
    TYPE_CHECKING,
    Tuple,
    Union,
)

from apischema.cache import cache
from apischema.conversions.utils import (
    Converter,
    converter_types,
    get_conversion_type,
    identity,
    is_convertible,
)
from apischema.dataclasses import replace
from apischema.types import AnyType
from apischema.utils import cached_property, is_method, method_class, method_wrapper

if TYPE_CHECKING:
    from apischema.deserialization.coercion import Coercion


@dataclass(frozen=True)
class Conversion:
    converter: Union[Converter, property]
    source: AnyType = None
    target: AnyType = None
    conversions: Optional["Conversions"] = None
    additional_properties: Optional[bool] = None
    coercion: Optional["Coercion"] = None
    default_fallback: Optional[bool] = None
    exclude_unset: Optional[bool] = None

    def __call__(self, *args, **kwargs):
        return self.converter(*args, **kwargs)


@dataclass(frozen=True)
class LazyConversion:
    get_conversion: Callable[[], Optional["Conversions"]]


ConvOrFunc = Union[Conversion, LazyConversion, Converter, property]
Conversions = Union[ConvOrFunc, Collection[ConvOrFunc]]
HashableConversions = Union[ConvOrFunc, Tuple[ConvOrFunc, ...]]


@dataclass
class ResolvedConversion:
    converter: Converter
    source: AnyType
    target: AnyType
    get_conversions: Optional[Callable[[], Optional[Conversions]]]
    additional_properties: Optional[bool]
    coercion: Optional["Coercion"]
    default_fallback: Optional[bool]
    exclude_unset: Optional[bool]

    @staticmethod
    def from_conversion(conversion: Conversion) -> "ResolvedConversion":
        assert not isinstance(conversion.converter, property)
        assert conversion.source is not None and conversion.target is not None
        get_conversions: Optional[Callable[[], Optional[Conversions]]]
        if conversion.conversions is None:
            get_conversions = None
        elif isinstance(conversion.conversions, LazyConversion):
            get_conversions = conversion.conversions.get_conversion  # type: ignore
        else:
            get_conversions = lambda: conversion.conversions  # noqa: E731
        return ResolvedConversion(
            conversion.converter,
            conversion.source,
            conversion.target,
            get_conversions,
            conversion.additional_properties,
            conversion.coercion,
            conversion.default_fallback,
            conversion.exclude_unset,
        )

    @cached_property
    def conversions(self) -> Optional["HashableConversions"]:
        if self.get_conversions is not None:
            return to_hashable_conversions(self.get_conversions())
        else:
            return None

    @property
    def is_identity(self) -> bool:
        return (
            self.converter == identity
            and self.source == self.target
            and all(
                attr is None
                for attr in (
                    self.get_conversions,
                    self.additional_properties,
                    self.coercion,
                    self.default_fallback,
                    self.exclude_unset,
                )
            )
        )


# ResolvedConversion = NewType("ResolvedConversion", Conversion)

ConversionResolver = Callable[[ConvOrFunc], ResolvedConversion]


def to_hashable_conversions(
    conversions: Optional[Conversions],
) -> Optional[HashableConversions]:
    if conversions is None:
        return None
    elif isinstance(conversions, Collection):
        return tuple(conversions)
    else:
        return conversions


def resolve_conversions(
    conversions: Conversions, resolver: ConversionResolver
) -> Collection[ResolvedConversion]:
    if not isinstance(conversions, Collection):
        conversions = [conversions]
    return tuple(map(resolver, conversions))


def to_conversion(conv: ConvOrFunc) -> Conversion:
    if isinstance(conv, LazyConversion):
        return conv.get_conversion()  # type: ignore
    elif not isinstance(conv, Conversion):
        return Conversion(conv)
    else:
        return conv


@cache
def resolve_deserialization(conversion: ConvOrFunc) -> ResolvedConversion:
    conversion = to_conversion(conversion)
    if isinstance(conversion.converter, property):
        raise TypeError("Properties cannot be used for deserialization conversions")
    source, target = converter_types(
        conversion.converter, conversion.source, conversion.target
    )
    target, source = get_conversion_type(target, source)
    if not is_convertible(target):
        raise TypeError(f"Target of {conversion} is not a convertible class")
    return ResolvedConversion.from_conversion(
        replace(conversion, source=source, target=target)
    )


def handle_serialization_method(conversion: ConvOrFunc) -> Conversion:
    conversion = to_conversion(conversion)
    if is_method(conversion.converter):
        if conversion.source is None:
            conversion = replace(conversion, source=method_class(conversion.converter))
        return replace(conversion, converter=method_wrapper(conversion.converter))
    else:
        return conversion


@cache
def resolve_serialization(conversion: ConvOrFunc) -> ResolvedConversion:
    conversion = handle_serialization_method(conversion)
    assert not isinstance(conversion.converter, property)
    source, target = converter_types(
        conversion.converter, conversion.source, conversion.target
    )
    source, target = get_conversion_type(source, target)
    if not is_convertible(source):
        raise TypeError(f"Source of {conversion} is not a convertible class")
    return ResolvedConversion.from_conversion(
        replace(conversion, source=source, target=target)
    )



from dataclasses import dataclass
from enum import Enum
from typing import (
    Any,
    Dict,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    cast,
)

from apischema.dataclass_utils import is_dataclass
from apischema.types import AnyType
from apischema.typing import _TypedDictMeta, get_origin
from apischema.utils import has_type_vars, is_type_var, replace_builtins, type_name
from apischema.visitor import Unsupported, Visitor

Ref = Union[str, "ellipsis", None]  # noqa: F821
_refs: Dict[AnyType, Optional[Ref]] = {}


def _default_ref(tp: AnyType) -> Ref:
    if not has_type_vars(tp) and (
        (
            isinstance(tp, type)
            and (
                is_dataclass(tp)
                or (issubclass(tp, tuple) and hasattr(tp, "_fields"))
                or issubclass(tp, Enum)
            )
        )
        or (hasattr(tp, "__supertype__") and is_builtin(tp))
        or isinstance(tp, _TypedDictMeta)
    ):
        return ...
    else:
        return None


def get_ref(tp: AnyType) -> Optional[str]:
    tp = replace_builtins(tp)
    ref = _refs[tp] if tp in _refs else _default_ref(tp)
    return cast(Optional[str], type_name(tp) if ref is ... else ref)


T = TypeVar("T")


@dataclass(frozen=True)
class schema_ref:
    ref: Ref = ...

    def __post_init__(self):
        if self.ref == "":
            raise ValueError("Empty schema ref not allowed")

    def check_type(self, tp: AnyType):
        if hasattr(tp, "__supertype__") and not is_builtin(tp):
            # NewType of non-builtin types cannot have a ref because their serialization
            # could be customized, but the NewType ref would then erase this
            # customization in the schema.
            raise TypeError("NewType of non-builtin type can not have a ref")
        if is_type_var(tp):
            raise TypeError("TypeVar cannot have a ref")
        elif has_type_vars(tp):
            raise TypeError("Unspecialized generic types cannot have a ref")
        if get_origin(tp) is not None and self.ref is ...:
            raise TypeError(f"Generic alias {tp} cannot have ... ref")

    def __call__(self, tp: T) -> T:
        self.check_type(tp)
        _refs[replace_builtins(tp)] = self.ref
        return tp


class BuiltinVisitor(Visitor):
    def collection(self, cls: Type[Iterable], value_type: AnyType):
        self.visit(value_type)

    def enum(self, cls: Type[Enum]):
        pass

    def literal(self, values: Sequence[Any]):
        pass

    def mapping(self, cls: Type[Mapping], key_type: AnyType, value_type: AnyType):
        self.visit(key_type), self.visit(value_type)

    def primitive(self, cls: Type):
        pass

    def subprimitive(self, cls: Type, superclass: Type):
        raise NotImplementedError

    def union(self, alternatives: Sequence[AnyType]):
        for alt in alternatives:
            self.visit(alt)


def is_builtin(tp: AnyType) -> bool:
    try:
        BuiltinVisitor().visit(tp)
    except (NotImplementedError, Unsupported):
        return False
    else:
        return True



from setuptools import find_packages, setup

with open("README.md") as f:
    README = f.read()

setup(
    name="apischema",
    version="0.14.1",
    url="https://github.com/wyfo/apischema",
    author="Joseph Perez",
    author_email="joperez@hotmail.fr",
    license="MIT",
    packages=find_packages(include=["apischema*"]),
    package_data={"apischema": ["py.typed"]},
    description="JSON (de)serialization + GraphQL and JSON schema generation through python typing, with a spoonful of sugar.",  # noqa: E501
    long_description=README,
    long_description_content_type="text/markdown",
    python_requires=">=3.6",
    install_requires=["dataclasses==0.7;python_version<'3.7'"],
    extras_require={"graphql": ["graphql-core>=3.1.2"]},
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



from collections import defaultdict
from dataclasses import dataclass, is_dataclass
from enum import Enum
from inspect import Parameter, iscoroutinefunction, signature
from typing import (
    Any,
    Awaitable,
    Callable,
    Collection,
    Dict,
    Iterator,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    cast,
    overload,
)

import graphql

from apischema.aliases import Aliaser
from apischema.conversions.conversions import Conversions, to_hashable_conversions
from apischema.conversions.dataclass_models import DataclassModel
from apischema.deserialization import deserialize
from apischema.json_schema.schema import Schema
from apischema.serialization import (
    COLLECTION_TYPE_SET,
    MAPPING_TYPE_SET,
    PRIMITIVE_TYPES_SET,
    get_conversions,
    serialize,
)
from apischema.serialization.serialized_methods import (
    ErrorHandler,
    Serialized,
    SerializedDescriptor,
    _get_methods,
    register_serialized,
)
from apischema.types import AnyType, NoneType, PRIMITIVE_TYPES
from apischema.typing import get_origin, get_type_hints
from apischema.utils import (
    MethodOrProperty,
    Undefined,
    get_args2,
    get_origin_or_type,
    is_method,
    is_union_of,
    method_class,
    method_wrapper,
)
from apischema.validation.errors import ValidationError
from apischema.visitor import Unsupported


def partial_serialize(
    obj: Any, *, aliaser: Aliaser, conversions: Conversions = None
) -> Any:
    if obj is Undefined:
        return None
    cls = obj.__class__
    if cls in PRIMITIVE_TYPES_SET:
        return obj
    if cls in COLLECTION_TYPE_SET:
        return [
            partial_serialize(elt, aliaser=aliaser, conversions=conversions)
            for elt in obj
        ]
    if cls in MAPPING_TYPE_SET:
        return serialize(
            obj, conversions=conversions, aliaser=aliaser, exclude_unset=False
        )
    conversion = get_conversions(cls, to_hashable_conversions(conversions))
    if conversion is not None:
        if isinstance(conversion.target, DataclassModel):
            return obj
        return partial_serialize(
            conversion.converter(obj),  # type: ignore
            aliaser=aliaser,
            conversions=conversion.conversions,
        )
    if is_dataclass(cls):
        return obj
    if issubclass(cls, Enum):
        return serialize(obj.value, aliaser=aliaser, exclude_unset=False)
    if isinstance(obj, PRIMITIVE_TYPES):
        return obj
    if isinstance(obj, Mapping):
        return serialize(obj, aliaser=aliaser, exclude_unset=False)
    if isinstance(obj, Collection):
        return [partial_serialize(elt, aliaser=aliaser) for elt in obj]
    if issubclass(cls, tuple) and hasattr(cls, "_fields"):
        return obj
    raise Unsupported(cls)


awaitable_origin = get_origin(Awaitable[Any])


def is_async(func: Callable) -> bool:
    return (
        iscoroutinefunction(func)
        or get_origin_or_type(get_type_hints(func).get("return")) == awaitable_origin
    )


def unwrap_awaitable(tp: AnyType) -> AnyType:
    return get_args2(tp)[0] if get_origin_or_type(tp) == awaitable_origin else tp


@dataclass(frozen=True)
class Resolver(Serialized):
    parameters: Sequence[Parameter]

    def error_type(self) -> AnyType:
        return unwrap_awaitable(super().error_type())

    def return_type(self, return_type: AnyType) -> AnyType:
        return super().return_type(unwrap_awaitable(return_type))


_resolvers: Dict[Type, Dict[str, Resolver]] = defaultdict(dict)


def get_resolvers(tp: AnyType) -> Mapping[str, Tuple[Resolver, Mapping[str, AnyType]]]:
    return _get_methods(tp, _resolvers)


def none_error_handler(
    __error: Exception, __obj: Any, __info: graphql.GraphQLResolveInfo, **kwargs
) -> None:
    return None


def resolver_parameters(
    resolver: Callable, *, check_first: bool
) -> Iterator[Parameter]:
    first = True
    for param in signature(resolver).parameters.values():
        if param.kind is Parameter.POSITIONAL_ONLY:
            raise TypeError("Resolver can not have positional only parameters")
        if param.kind in {Parameter.POSITIONAL_OR_KEYWORD, Parameter.KEYWORD_ONLY}:
            if param.annotation is Parameter.empty and (check_first or not first):
                raise TypeError("Resolver parameters must be typed")
            yield param
        first = False


def register_resolver(
    func: Callable,
    alias: str,
    conversions: Optional[Conversions],
    schema: Optional[Schema],
    error_handler: ErrorHandler,
    serialized: bool,
    owner: Type = None,
):
    first_param, *parameters = resolver_parameters(func, check_first=owner is None)
    if error_handler is None:
        error_handler = none_error_handler
    elif error_handler is Undefined:
        error_handler = None
    resolver = Resolver(func, conversions, schema, error_handler, parameters)
    if owner is None:
        try:
            owner = get_origin_or_type(get_type_hints(func)[first_param.name])
        except KeyError:
            raise TypeError("First parameter of resolver must be typed") from None
    _resolvers[owner][alias] = resolver
    if serialized:
        if is_async(func):
            raise TypeError("Async resolver cannot be used as a serialized method")
        try:
            register_serialized(func, alias, conversions, schema, error_handler, owner)
        except Exception:
            raise TypeError("Resolver cannot be used as a serialized method")


class ResolverDescriptor(SerializedDescriptor):
    def __init__(
        self,
        method: MethodOrProperty,
        alias: Optional[str],
        conversions: Optional[Conversions],
        schema: Optional[Schema],
        error_handler: ErrorHandler,
        serialized: bool,
    ):
        super().__init__(method, alias, conversions, schema, error_handler)
        self.serialized = serialized

    def __set_name__(self, owner, name):
        register_resolver(
            method_wrapper(self._method),
            self._alias or name,
            self._conversions,
            self._schema,
            self._error_handler,
            self.serialized,
            owner,
        )
        setattr(owner, name, self._method)


MethodOrProp = TypeVar("MethodOrProp", Callable, property)


@overload
def resolver(__method_or_property: MethodOrProp) -> MethodOrProp:
    ...


@overload
def resolver(
    alias: str = None,
    *,
    conversions: Conversions = None,
    schema: Schema = None,
    error_handler: ErrorHandler = Undefined,
    serialized: bool = False,
    owner: Type = None,
) -> Callable[[MethodOrProp], MethodOrProp]:
    ...


def resolver(
    __arg=None,
    *,
    alias: str = None,
    conversions: Conversions = None,
    schema: Schema = None,
    error_handler: ErrorHandler = Undefined,
    serialized: bool = False,
    owner: Type = None,
):
    def decorator(method: MethodOrProperty):
        nonlocal owner
        if is_method(method) and method_class(method) is None:
            return ResolverDescriptor(
                method, alias, conversions, schema, error_handler, serialized
            )
        else:
            if is_method(method):
                if owner is None:
                    owner = method_class(method)
                method = method_wrapper(method)
            assert not isinstance(method, property)
            register_resolver(
                cast(Callable, method),
                alias or method.__name__,
                conversions,
                schema,
                error_handler,
                serialized,
                owner,
            )
            return method

    if isinstance(__arg, str) or __arg is None:
        alias = alias or __arg
        return decorator
    else:
        return decorator(__arg)


def resolver_resolve(
    resolver: Resolver,
    types: Mapping[str, AnyType],
    aliaser: Aliaser,
    serialized: bool = True,
) -> Callable:
    parameters, info_parameter = [], None
    for param in resolver.parameters:
        param_type = types[param.name]
        if is_union_of(param_type, graphql.GraphQLResolveInfo):
            info_parameter = param.name
        else:
            parameters.append(
                (
                    aliaser(param.name),
                    param.name,
                    param_type,
                    is_union_of(param_type, NoneType),
                    param.default is Parameter.empty,
                )
            )
    func, error_handler = resolver.func, resolver.error_handler

    def no_serialize(result):
        return result

    async def async_serialize(result: Awaitable):
        return partial_serialize(
            await result, aliaser=aliaser, conversions=resolver.conversions
        )

    def sync_serialize(result):
        return partial_serialize(
            result, aliaser=aliaser, conversions=resolver.conversions
        )

    serialize_result: Callable[[Any], Any]
    if not serialized:
        serialize_result = no_serialize
    elif is_async(resolver.func):
        serialize_result = async_serialize
    else:
        serialize_result = sync_serialize
    serialize_error: Optional[Callable[[Any], Any]]
    if error_handler is None:
        serialize_error = None
    elif is_async(error_handler):
        serialize_error = async_serialize
    else:
        serialize_error = sync_serialize

    def resolve(__self, __info, **kwargs):
        values = {}
        errors: Dict[str, ValidationError] = {}
        for alias, param_name, param_type, opt_param, is_required in parameters:
            if param_name in kwargs:
                if not opt_param and kwargs[param_name] is None:
                    assert not is_required
                    continue
                try:
                    values[param_name] = deserialize(
                        param_type, kwargs[param_name], aliaser=aliaser
                    )
                except ValidationError as err:
                    errors[aliaser(param_name)] = err
            elif opt_param and is_required:
                values[param_name] = None

        if errors:
            raise TypeError(serialize(ValidationError(children=errors)))
        if info_parameter:
            values[info_parameter] = __info
        try:
            return serialize_result(func(__self, **values))
        except Exception as error:
            if error_handler is None:
                raise
            assert serialize_error is not None
            return serialize_error(error_handler(error, __self, __info, **kwargs))

    return resolve



from dataclasses import Field, InitVar, dataclass, field as field_
from enum import Enum
from inspect import Parameter, iscoroutinefunction
from typing import (
    Any,
    AsyncIterable,
    AsyncIterator,
    Callable,
    Collection,
    Dict,
    Generic,
    Iterable,
    List,
    Mapping,
    NewType,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

import graphql

from apischema import UndefinedType, serialize
from apischema.aliases import Aliaser
from apischema.conversions import identity
from apischema.conversions.conversions import Conversions
from apischema.conversions.utils import Converter
from apischema.conversions.visitor import (
    Conv,
    ConversionsVisitor,
    Deserialization,
    DeserializationVisitor,
    Serialization,
    SerializationVisitor,
)
from apischema.dataclass_utils import (
    get_alias,
    get_default,
    get_field_conversion,
    get_fields,
    is_required,
)
from apischema.dataclasses import replace
from apischema.graphql.interfaces import get_interfaces, is_interface
from apischema.graphql.resolvers import (
    Resolver,
    get_resolvers,
    none_error_handler,
    partial_serialize,
    resolver_parameters,
    resolver_resolve,
)
from apischema.json_schema.refs import get_ref, schema_ref
from apischema.json_schema.schema import Schema, get_schema, merge_schema
from apischema.metadata.keys import (
    MERGED_METADATA,
    PROPERTIES_METADATA,
    SCHEMA_METADATA,
    check_metadata,
)
from apischema.serialization.serialized_methods import ErrorHandler
from apischema.skip import filter_skipped
from apischema.types import AnyType, NoneType
from apischema.typing import get_origin
from apischema.utils import (
    Undefined,
    get_args2,
    get_origin2,
    is_union_of,
    to_camel_case,
    type_name,
)

JsonScalar = graphql.GraphQLScalarType(
    "JSON",
    specified_by_url="http://www.ecma-international.org/publications/files/ECMA-ST/ECMA-404.pdf",  # noqa: E501
)
GRAPHQL_PRIMITIVE_TYPES = {
    int: graphql.GraphQLInt,
    float: graphql.GraphQLFloat,
    str: graphql.GraphQLString,
    bool: graphql.GraphQLBoolean,
}

ID = NewType("ID", str)


class MissingRef(Exception):
    pass


class Nullable(Exception):
    pass


def ref_or_name(tp: AnyType) -> str:
    return get_ref(tp) or type_name(tp)


T = TypeVar("T")
Lazy = Callable[[], T]
Thunk = Union[Callable[[], T], T]


def exec_thunk(thunk: Thunk[graphql.GraphQLType], *, non_null=None) -> Any:
    result = thunk if isinstance(thunk, graphql.GraphQLType) else thunk()
    if non_null is True and not isinstance(result, graphql.GraphQLNonNull):
        return graphql.GraphQLNonNull(result)  # type: ignore
    if non_null is False and isinstance(result, graphql.GraphQLNonNull):
        return result.of_type
    return result


@dataclass(frozen=True)
class ObjectField:
    name: str
    type: AnyType
    alias: Optional[str] = None
    converter: Converter = identity
    conversions: Optional[Conversions] = None
    default: Any = graphql.Undefined
    parameters: Optional[Tuple[Collection[Parameter], Mapping[str, AnyType]]] = None
    resolve: Optional[Callable] = None
    schema: InitVar[Optional[Schema]] = None
    subscribe: Optional[Callable] = None
    deprecated: Optional[str] = field_(init=False, default=None)
    description: Optional[str] = field_(init=False, default=None)

    def __post_init__(self, schema: Optional[Schema]):
        if schema is not None and schema.annotations is not None:
            object.__setattr__(self, "description", schema.annotations.description)
            if schema.annotations.deprecated is True:
                object.__setattr__(
                    self, "deprecated", graphql.DEFAULT_DEPRECATION_REASON
                )
            elif isinstance(schema.annotations.deprecated, str):
                object.__setattr__(self, "deprecated", schema.annotations.deprecated)
            if schema.annotations.default is not Undefined:
                object.__setattr__(self, "default", schema.annotations.default)


IdPredicate = Callable[[AnyType], bool]
UnionRefFactory = Callable[[Sequence[str]], str]


class SchemaBuilder(ConversionsVisitor[Conv, Thunk[graphql.GraphQLType]]):
    def __init__(
        self,
        aliaser: Optional[Aliaser],
        id_type: graphql.GraphQLScalarType,
        is_id: Optional[IdPredicate],
        union_ref_factory: Optional[UnionRefFactory],
    ):
        super().__init__()
        self.aliaser = aliaser or (lambda s: s)
        self.id_type = id_type
        self.is_id = is_id or (lambda t: False)
        self.union_ref_factory = union_ref_factory
        self._cache: Dict[Any, Thunk[graphql.GraphQLType]] = {}
        self._non_null = True
        self._ref: Optional[str] = None
        self._schema: Optional[Schema] = None

    @property
    def _description(self) -> Optional[str]:
        if self._schema is not None and self._schema.annotations is not None:
            return self._schema.annotations.description
        else:
            return None

    @property
    def _ref_and_desc(self) -> Tuple[str, Optional[str]]:
        if self._ref is None:
            raise MissingRef
        return self._ref, self._description

    def annotated(
        self, tp: AnyType, annotations: Sequence[Any]
    ) -> Thunk[graphql.GraphQLType]:
        for annotation in annotations:
            if isinstance(annotation, schema_ref):
                annotation.check_type(tp)
                ref = annotation.ref
                if not isinstance(ref, str):
                    raise ValueError("Annotated schema_ref can only be str")
                self._ref = self._ref or ref
            if isinstance(annotation, Schema):
                self._schema = merge_schema(annotation, self._schema)
        return self.visit_with_schema(tp, self._ref, self._schema)

    def any(self) -> Thunk[graphql.GraphQLType]:
        return JsonScalar

    def collection(
        self, cls: Type[Collection], value_type: AnyType
    ) -> Thunk[graphql.GraphQLType]:
        value_thunk = self.visit(value_type)
        return lambda: graphql.GraphQLList(exec_thunk(value_thunk))

    def _object_field(self, field: Field, field_type: AnyType) -> ObjectField:
        field_name = field.name
        field_type, conversion = get_field_conversion(field, field_type, self.operation)
        converter, conversions = identity, None
        if conversion is not None:
            converter = conversion.converter  # type: ignore
            conversions = conversion.conversions
        return ObjectField(
            field_name,
            field_type,
            alias=get_alias(field),
            converter=converter,
            conversions=conversions,
            default=graphql.Undefined if is_required(field) else get_default(field),
            schema=field.metadata.get(SCHEMA_METADATA),
        )

    def _visit_merged(
        self, field: Field, field_type: AnyType
    ) -> Thunk[graphql.GraphQLType]:
        field_type, conversion = get_field_conversion(field, field_type, self.operation)
        return self.visit_with_conversions(
            field_type, conversion.conversions if conversion is not None else None
        )

    def dataclass(
        self,
        cls: Type,
        types: Mapping[str, AnyType],
        fields: Sequence[Field],
        init_vars: Sequence[Field],
    ) -> Thunk[graphql.GraphQLType]:
        types = dict(types)
        object_fields: List[ObjectField] = []
        merged_types: Dict[str, Thunk[graphql.GraphQLType]] = {}
        for field in get_fields(fields, init_vars, self.operation):
            check_metadata(field)
            if MERGED_METADATA in field.metadata:
                merged_types[field.name] = self._visit_merged(field, types[field.name])
            elif PROPERTIES_METADATA in field.metadata:
                continue
            else:
                object_fields.append(self._object_field(field, types[field.name]))
        return self.object(cls, object_fields, merged_types)

    def enum(self, cls: Type[Enum]) -> Thunk[graphql.GraphQLType]:
        return self.literal([elt.value for elt in cls])

    def generic(self, tp: AnyType) -> Thunk[graphql.GraphQLType]:
        self._ref = self._ref or get_ref(tp)
        return super().generic(tp)

    def literal(self, values: Sequence[Any]) -> Thunk[graphql.GraphQLType]:
        if not all(isinstance(v, str) for v in values):
            raise TypeError("Apischema GraphQL only support Enum/Literal of strings")
        name, description = self._ref_and_desc
        return graphql.GraphQLEnumType(
            name, dict(zip(values, values)), description=description
        )

    def mapping(
        self, cls: Type[Mapping], key_type: AnyType, value_type: AnyType
    ) -> Thunk[graphql.GraphQLType]:
        try:
            name, description = self._ref_and_desc
            return graphql.GraphQLScalarType(name, description=description)
        except MissingRef:
            return JsonScalar

    def named_tuple(
        self,
        cls: Type[Tuple],
        types: Mapping[str, AnyType],
        defaults: Mapping[str, Any],
    ) -> Thunk[graphql.GraphQLType]:
        fields = [
            ObjectField(
                field_name,
                field_type,
                default=defaults.get(field_name, graphql.Undefined),
            )
            for field_name, field_type in types.items()
        ]
        return self.object(cls, fields)

    def new_type(self, tp: Type, super_type: AnyType) -> Thunk[graphql.GraphQLType]:
        return self.visit_with_schema(super_type, self._ref, self._schema)

    def object(
        self,
        cls: Type,
        fields: Collection[ObjectField],
        merged_types: Mapping[str, Thunk[graphql.GraphQLType]] = None,
    ) -> Thunk[graphql.GraphQLType]:
        raise NotImplementedError

    def primitive(self, cls: Type) -> Thunk[graphql.GraphQLType]:
        if cls is NoneType:
            raise Nullable
        try:
            name, description = self._ref_and_desc
            return graphql.GraphQLScalarType(name, description=description)
        except MissingRef:
            return GRAPHQL_PRIMITIVE_TYPES[cls]

    def tuple(self, types: Sequence[AnyType]) -> Thunk[graphql.GraphQLType]:
        raise TypeError("Tuple are not supported")

    def union(self, alternatives: Sequence[AnyType]) -> Thunk[graphql.GraphQLType]:
        alternatives = list(filter_skipped(alternatives, schema_only=True))
        results = []
        for alt in alternatives:
            try:
                results.append(self.visit(alt))
            except Nullable:
                self._non_null = False
        if not results:
            raise TypeError("Empty union")
        return self._union_result(results)

    def visit_with_schema(
        self, tp: AnyType, ref: Optional[str], schema: Optional[Schema]
    ) -> Thunk[graphql.GraphQLType]:
        if self.is_id(tp) or tp == ID:
            return graphql.GraphQLNonNull(self.id_type)
        if self._apply_dynamic_conversions(tp) is None:
            ref, schema = ref or get_ref(tp), merge_schema(get_schema(tp), schema)
        else:
            ref, schema = None, None
        ref_save, schema_save, non_null_save = self._ref, self._schema, self._non_null
        self._ref, self._schema, self._non_null = ref, schema, True
        try:
            result = super().visit(tp)
            non_null = self._non_null
            return lambda: exec_thunk(result, non_null=non_null)
        except MissingRef:
            raise TypeError(f"Missing ref for type {tp}") from None
        finally:
            self._ref, self._schema = ref_save, schema_save
            self._non_null = non_null_save

    def _visit(self, tp: AnyType) -> Thunk[graphql.GraphQLType]:
        key = self._generic or tp, self._ref, self._schema, self._conversions
        if key in self._cache:
            return self._cache[key]
        cache = None

        def rec_sentinel() -> graphql.GraphQLType:
            assert cache is not None
            return cache

        self._cache[key] = rec_sentinel
        try:
            cache = exec_thunk(super()._visit(tp))
        except Exception:
            del self._cache[key]
            raise
        else:
            return cache

    def visit(self, tp: AnyType) -> Thunk[graphql.GraphQLType]:
        return self.visit_with_schema(tp, None, None)


FieldType = TypeVar("FieldType", graphql.GraphQLInputField, graphql.GraphQLField)


def merge_fields(
    cls: Type,
    fields: Mapping[str, Lazy[FieldType]],
    merged_types: Mapping[str, Thunk[graphql.GraphQLType]],
) -> Dict[str, FieldType]:
    all_merged_fields: Dict[str, FieldType] = {}
    for merged_name, merged_thunk in merged_types.items():
        merged_type = exec_thunk(merged_thunk, non_null=False)
        if not isinstance(
            merged_type,
            (
                graphql.GraphQLObjectType,
                graphql.GraphQLInterfaceType,
                graphql.GraphQLInputObjectType,
            ),
        ):
            raise TypeError(
                f"Merged field {cls.__name__}.{merged_name} must have an object type"
            )
        merged_fields: Mapping[str, FieldType] = merged_type.fields
        if merged_fields.keys() & all_merged_fields.keys() & fields.keys():
            raise TypeError(f"Conflict in merged fields of {cls}")
        all_merged_fields.update(merged_fields)
    return {**{name: field() for name, field in fields.items()}, **all_merged_fields}


class InputSchemaBuilder(
    DeserializationVisitor[Thunk[graphql.GraphQLType]],
    SchemaBuilder[Deserialization],
):
    def _field(self, field: ObjectField) -> Tuple[str, Lazy[graphql.GraphQLInputField]]:
        field_type = field.type
        default: Any = graphql.Undefined
        # Don't put `null` default + handle Undefined as None
        if field.default in {None, Undefined}:
            field_type = Optional[field_type]
        elif field.default is not graphql.Undefined:
            try:
                default = serialize(
                    field.converter(field.default),
                    conversions=field.conversions,
                    aliaser=self.aliaser,
                )
            except Exception:
                field_type = Optional[field_type]

        type_thunk = self.visit_with_conversions(field_type, field.conversions)

        def field_thunk():
            return graphql.GraphQLInputField(
                exec_thunk(type_thunk),
                default_value=default,
                description=field.description,
            )

        return self.aliaser(field.alias or field.name), field_thunk

    def object(
        self,
        cls: Type,
        fields: Collection[ObjectField],
        merged_types: Mapping[str, Thunk[graphql.GraphQLType]] = None,
    ) -> Thunk[graphql.GraphQLType]:
        name, description = self._ref_and_desc
        name = name if name.endswith("Input") else name + "Input"
        visited_fields = dict(map(self._field, fields))
        return lambda: graphql.GraphQLInputObjectType(
            name,
            lambda: merge_fields(cls, visited_fields, merged_types or {}),
            description,
        )

    def typed_dict(
        self, cls: Type, keys: Mapping[str, AnyType], total: bool
    ) -> Thunk[graphql.GraphQLType]:
        return self.object(
            cls, [ObjectField(name, type) for name, type in keys.items()]
        )

    def _union_result(
        self, results: Iterable[Thunk[graphql.GraphQLType]]
    ) -> Thunk[graphql.GraphQLType]:
        results = list(results)  # Execute the iteration
        if len(results) == 1:
            return results[0]
        raise TypeError("Union are not supported for input")


class OutputSchemaBuilder(
    SerializationVisitor[Thunk[graphql.GraphQLType]], SchemaBuilder[Serialization]
):
    def __init__(
        self,
        aliaser: Optional[Aliaser],
        id_type: graphql.GraphQLScalarType,
        is_id: Optional[IdPredicate],
        union_ref_factory: Optional[UnionRefFactory],
    ):
        super().__init__(aliaser, id_type, is_id, union_ref_factory)
        self.input_builder = InputSchemaBuilder(
            aliaser, id_type, is_id, union_ref_factory
        )
        self._get_merged: Optional[Callable] = None

    def _wrap_resolve(self, resolve: Callable):
        if self._get_merged is None:
            return resolve
        else:
            get_merged = self._get_merged

            def resolve_wrapper(__obj, __info, **kwargs):
                return resolve(get_merged(__obj), __info, **kwargs)

            return resolve_wrapper

    def _field(self, field: ObjectField) -> Tuple[str, Lazy[graphql.GraphQLField]]:
        if field.resolve is not None:
            resolve = field.resolve
        else:
            field_name, aliaser = field.name, self.aliaser
            converter, conversions = field.converter, field.conversions

            def resolve(obj, _):
                return partial_serialize(
                    converter(getattr(obj, field_name)),
                    aliaser=aliaser,
                    conversions=conversions,
                )

        resolve = self._wrap_resolve(resolve)

        field_type = field.type
        if is_union_of(field_type, UndefinedType):
            field_type = Optional[field_type]
        type_thunk = self.visit_with_conversions(field_type, field.conversions)
        args = None
        if field.parameters is not None:
            parameters, types = field.parameters
            args = {}
            for param in parameters:
                default: Any = graphql.Undefined
                param_type = types[param.name]
                if is_union_of(param_type, graphql.GraphQLResolveInfo):
                    break
                # Don't put `null` default + handle Undefined as None
                # also https://github.com/python/typing/issues/775
                if param.default in {None, Undefined}:
                    param_type = Optional[param_type]
                # param.default == graphql.Undefined means the parameter is required
                # even if it has a default
                elif param.default not in {Parameter.empty, graphql.Undefined}:
                    try:
                        default = serialize(param.default)
                    except Exception:
                        param_type = Optional[param_type]

                def arg_thunk(
                    arg_thunk=self.input_builder.visit(param_type),
                    default=default,
                ) -> graphql.GraphQLArgument:
                    return graphql.GraphQLArgument(exec_thunk(arg_thunk), default)

                args[self.aliaser(param.name)] = arg_thunk
        return self.aliaser(field.alias or field.name), lambda: graphql.GraphQLField(
            exec_thunk(type_thunk),
            {name: arg() for name, arg in args.items()} if args else None,
            resolve,
            field.subscribe,
            field.description,
            field.deprecated,
        )

    def _visit_merged(
        self, field: Field, field_type: AnyType
    ) -> Thunk[graphql.GraphQLType]:
        field_type, conversion = get_field_conversion(field, field_type, self.operation)
        field_name, aliaser = field.name, self.aliaser
        converter: Converter = identity
        conversions: Optional[Conversions] = None
        if conversion is not None:
            converter = conversion.converter  # type: ignore
            conversions = conversion.conversions
        get_prev_merged = self._get_merged if self._get_merged is not None else identity

        def get_merge(obj):
            return partial_serialize(
                converter(getattr(get_prev_merged(obj), field_name)),
                aliaser=aliaser,
                conversions=conversions,
            )

        merged_save = self._get_merged
        self._get_merged = get_merge
        try:
            return self.visit_with_conversions(
                field_type,
                conversion.conversions if conversion is not None else None,
            )
        finally:
            self._get_merged = merged_save

    def object(
        self,
        cls: Type,
        fields: Collection[ObjectField],
        merged_types: Mapping[str, Thunk[graphql.GraphQLType]] = None,
    ) -> Thunk[graphql.GraphQLType]:
        fields_and_resolvers = list(fields)
        try:
            name, description = self._ref_and_desc
        except MissingRef:
            if cls.__name__ not in ("Query", "Mutation", "Subscription"):
                raise
            name, description = cls.__name__, self._description
        for resolver_name, (resolver, types) in get_resolvers(
            self._generic or cls
        ).items():
            resolver_field = ObjectField(
                resolver_name,
                types["return"],
                conversions=resolver.conversions,
                parameters=(resolver.parameters, types),
                resolve=self._wrap_resolve(
                    resolver_resolve(resolver, types, self.aliaser)
                ),
            )
            fields_and_resolvers.append(resolver_field)
        visited_fields = dict(map(self._field, fields_and_resolvers))

        def field_thunk() -> graphql.GraphQLFieldMap:
            return merge_fields(
                cls,
                visited_fields,
                merged_types or {},
            )

        interfaces = list(map(self.visit, get_interfaces(cls)))
        interface_thunk = None
        if interfaces:

            def interface_thunk() -> Collection[graphql.GraphQLInterfaceType]:
                result = {exec_thunk(i, non_null=False) for i in interfaces}
                for merged_thunk in (merged_types or {}).values():
                    merged = cast(
                        Union[graphql.GraphQLObjectType, graphql.GraphQLInterfaceType],
                        exec_thunk(merged_thunk, non_null=False),
                    )
                    result.update(merged.interfaces)
                return sorted(result, key=lambda i: i.name)

        if is_interface(cls):
            return lambda: graphql.GraphQLInterfaceType(
                name,
                field_thunk,
                interface_thunk,
                description=description,
            )

        else:
            return lambda: graphql.GraphQLObjectType(
                name,
                field_thunk,
                interface_thunk,
                is_type_of=lambda obj, _: isinstance(obj, cls),
                description=description,
            )

    def typed_dict(
        self, cls: Type, keys: Mapping[str, AnyType], total: bool
    ) -> Thunk[graphql.GraphQLType]:
        raise TypeError("TyedDict are not supported in output schema")

    def _union_result(
        self, results: Iterable[Thunk[graphql.GraphQLType]]
    ) -> Thunk[graphql.GraphQLType]:
        results = list(results)  # Execute the iteration
        if len(results) == 1:
            return results[0]
        name, description = self._ref, self._description
        if name is None and self.union_ref_factory is None:
            raise MissingRef

        def thunk() -> graphql.GraphQLUnionType:
            types = [exec_thunk(res, non_null=False) for res in results]
            if name is None:
                assert self.union_ref_factory is not None
                computed_name = self.union_ref_factory([t.name for t in types])
            else:
                computed_name = name
            return graphql.GraphQLUnionType(
                computed_name, types, description=description
            )

        return thunk


async_iterable_origins = set(map(get_origin, (AsyncIterable[Any], AsyncIterator[Any])))

_fake_type = cast(type, ...)


@dataclass(frozen=True)
class Operation(Generic[T]):
    function: Callable[..., T]
    alias: Optional[str] = None
    conversions: Optional[Conversions] = None
    schema: Optional[Schema] = None
    error_handler: ErrorHandler = Undefined


class Query(Operation):
    pass


class Mutation(Operation):
    pass


@dataclass(frozen=True)
class Subscription(Operation[AsyncIterable]):
    resolver: Optional[Callable] = None


Op = TypeVar("Op", bound=Operation)


def operation_resolver(
    operation: Union[Callable, Op], op_class: Type[Op]
) -> Tuple[str, Resolver]:
    if not isinstance(operation, op_class):
        operation = op_class(operation)  # type: ignore
    error_handler: Optional[Callable]
    if operation.error_handler is Undefined:
        error_handler = None
    elif operation.error_handler is None:
        error_handler = none_error_handler
    else:
        error_handler = operation.error_handler
    op = operation.function
    if iscoroutinefunction(op):

        async def wrapper(_, *args, **kwargs):
            return await op(*args, **kwargs)

    else:

        def wrapper(_, *args, **kwargs):
            return op(*args, **kwargs)

    wrapper.__annotations__ = op.__annotations__

    (*parameters,) = resolver_parameters(operation.function, check_first=True)
    return operation.alias or operation.function.__name__, Resolver(
        wrapper, operation.conversions, operation.schema, error_handler, parameters
    )


def graphql_schema(
    *,
    query: Iterable[Union[Callable, Query]] = (),
    mutation: Iterable[Union[Callable, Mutation]] = (),
    subscription: Iterable[Union[Callable[..., AsyncIterable], Subscription]] = (),
    types: Iterable[Type] = (),
    directives: Optional[Collection[graphql.GraphQLDirective]] = None,
    description: Optional[str] = None,
    extensions: Optional[Dict[str, Any]] = None,
    aliaser: Aliaser = to_camel_case,
    id_types: Union[Collection[AnyType], IdPredicate] = None,
    id_encoding: Tuple[
        Optional[Callable[[str], Any]], Optional[Callable[[Any], str]]
    ] = (None, None),
    union_ref: UnionRefFactory = "Or".join,
) -> graphql.GraphQLSchema:

    query_fields: List[ObjectField] = []
    mutation_fields: List[ObjectField] = []
    subscription_fields: List[ObjectField] = []
    for operations, op_class, fields in [
        (query, Query, query_fields),
        (mutation, Mutation, mutation_fields),
    ]:
        for operation in operations:  # type: ignore
            name, resolver = operation_resolver(operation, op_class)
            resolver_types = resolver.types()
            fields.append(
                ObjectField(
                    name,
                    resolver_types["return"],
                    conversions=resolver.conversions,
                    parameters=(resolver.parameters, resolver_types),
                    resolve=resolver_resolve(resolver, resolver_types, aliaser),
                    schema=resolver.schema,
                )
            )
    for sub_op in subscription:  # type: ignore
        if not isinstance(sub_op, Subscription):
            sub_op = Subscription(sub_op)  # type: ignore
        sub_parameters: Sequence[Parameter]
        if sub_op.resolver is not None:
            name = sub_op.alias or sub_op.resolver.__name__
            _, subscriber2 = operation_resolver(sub_op, Subscription)
            _, *sub_parameters = resolver_parameters(sub_op.resolver, check_first=False)
            resolver = Resolver(
                sub_op.resolver,
                sub_op.conversions,
                sub_op.schema,
                subscriber2.error_handler,
                sub_parameters,
            )
            sub_types = resolver.types()
            sub_return = sub_types["return"]
            resolve = resolver_resolve(resolver, sub_types, aliaser)
            subscriber = replace(subscriber2, error_handler=None)
            subscribe = resolver_resolve(
                subscriber, subscriber.types(), aliaser, serialized=False
            )
        else:
            name, subscriber2 = operation_resolver(sub_op, Subscription)
            resolver = Resolver(
                lambda _: _,
                sub_op.conversions,
                sub_op.schema,
                subscriber2.error_handler,
                (),
            )
            resolve = resolver_resolve(resolver, {}, aliaser)
            subscriber = replace(subscriber2, error_handler=None)
            sub_parameters = subscriber.parameters
            sub_types = subscriber.types()
            if get_origin2(sub_types["return"]) not in async_iterable_origins:
                raise TypeError(
                    "Subscriptions must return an AsyncIterable/AsyncIterator"
                )
            event_type = get_args2(sub_types["return"])[0]
            subscribe = resolver_resolve(
                subscriber, sub_types, aliaser, serialized=False
            )
            sub_return = resolver.return_type(event_type)

        subscription_fields.append(
            ObjectField(
                name,
                sub_return,
                conversions=sub_op.conversions,
                parameters=(sub_parameters, sub_types),
                resolve=resolve,
                subscribe=subscribe,
                schema=sub_op.schema,
            )
        )

    is_id = id_types.__contains__ if isinstance(id_types, Collection) else id_types
    if id_encoding == (None, None):
        id_type: graphql.GraphQLScalarType = graphql.GraphQLID
    else:
        id_deserializer, id_serializer = id_encoding
        id_type = graphql.GraphQLScalarType(
            name="ID",
            serialize=id_serializer or graphql.GraphQLID.serialize,
            parse_value=id_deserializer or graphql.GraphQLID.parse_value,
            parse_literal=graphql.GraphQLID.parse_literal,
            description=graphql.GraphQLID.description,
        )
    builder = OutputSchemaBuilder(aliaser, id_type, is_id, union_ref)

    def root_type(
        name: str, fields: Collection[ObjectField]
    ) -> Optional[graphql.GraphQLObjectType]:
        if not fields and name != "Query":
            return None
        return exec_thunk(builder.object(type(name, (), {}), fields), non_null=False)

    return graphql.GraphQLSchema(
        query=root_type("Query", query_fields),
        mutation=root_type("Mutation", mutation_fields),
        subscription=root_type("Subscription", subscription_fields),
        types=[exec_thunk(builder.visit(cls), non_null=False) for cls in types],
        directives=directives,
        description=description,
        extensions=extensions,
    )



__all__ = ["NotNull", "Skip"]
from typing import Iterator, Sequence, TypeVar, Union

from apischema.types import AnyType
from apischema.typing import get_args, get_origin
from apischema.utils import UndefinedType


class Skipped(Exception):
    pass


class _Skip:
    def __call__(self, *, schema_only: bool):
        return SkipSchema if schema_only else self


Skip = _Skip()
SkipSchema = object()

T = TypeVar("T")
try:
    from typing import Annotated  # type: ignore
except ImportError:
    try:
        from typing_extensions import Annotated  # type: ignore
    except ImportError:
        Annotated = None  # type: ignore

if Annotated is not None:
    NotNull = Union[T, Annotated[None, Skip]]
else:

    class _NotNull:
        def __getitem__(self, item):
            raise TypeError("NotNull requires Annotated (PEP 593)")

    NotNull = _NotNull()  # type: ignore


def is_skipped(tp: AnyType, *, schema_only) -> bool:
    return tp is UndefinedType or (
        get_origin(tp) is Annotated
        and (
            Skip in get_args(tp)[1:] or (schema_only and SkipSchema in get_args(tp)[1:])
        )
    )


def filter_skipped(
    alternatives: Sequence[AnyType], *, schema_only=False
) -> Iterator[AnyType]:
    return (alt for alt in alternatives if not is_skipped(alt, schema_only=schema_only))



from enum import Enum
from functools import wraps
from inspect import signature
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    Mapping,
    Pattern,
    Sequence,
    Type,
    TypeVar,
    Union,
    cast,
)


from apischema.types import NoneType, Number
from apischema.utils import Undefined


class JsonType(str, Enum):
    NULL = "null"
    BOOLEAN = "boolean"
    STRING = "string"
    INTEGER = "integer"
    NUMBER = "number"
    ARRAY = "array"
    OBJECT = "object"

    @staticmethod
    def from_type(cls: Type) -> "JsonType":
        try:
            return {
                NoneType: JsonType.NULL,
                bool: JsonType.BOOLEAN,
                str: JsonType.STRING,
                int: JsonType.INTEGER,
                float: JsonType.NUMBER,
                list: JsonType.ARRAY,
                dict: JsonType.OBJECT,
            }[cls]
        except KeyError:
            raise TypeError(f"Invalid JSON type {cls}")


class JsonSchema(Dict[str, Any]):
    pass


Func = TypeVar("Func", bound=Callable)


def json_schema_kwargs(func: Func) -> Func:
    @wraps(func)
    def wrapper(**kwargs):
        type_ = kwargs.get("type")
        if isinstance(type_, Sequence):
            if JsonType.INTEGER in type_ and JsonType.NUMBER in type_:
                kwargs["type"] = [t for t in type_ if t != JsonType.INTEGER]
        return JsonSchema(
            (k, v)
            for k, v in kwargs.items()
            if k not in _json_schema_params or v != _json_schema_params[k].default
        )

    _json_schema_params = signature(func).parameters
    return cast(Func, wrapper)


@json_schema_kwargs
def json_schema(
    *,
    additionalProperties: Union[bool, JsonSchema] = JsonSchema(),
    allOf: Sequence[JsonSchema] = [],
    anyOf: Sequence[JsonSchema] = [],
    const: Any = Undefined,
    default: Any = Undefined,
    dependentRequired: Mapping[str, Collection[str]] = {},
    deprecated: bool = False,
    description: str = None,
    enum: Sequence[Any] = [],
    exclusiveMaximum: Number = None,
    exclusiveMinimum: Number = None,
    examples: Sequence[Any] = None,
    format: str = None,
    items: Union[JsonSchema, Sequence[JsonSchema]] = JsonSchema(),
    maximum: Number = None,
    minimum: Number = None,
    maxItems: int = None,
    minItems: int = None,
    maxLength: int = None,
    minLength: int = None,
    maxProperties: int = None,
    minProperties: int = None,
    multipleOf: Number = None,
    oneOf: Sequence[JsonSchema] = [],
    pattern: Pattern = None,
    patternProperties: Mapping[Pattern, JsonSchema] = {},
    properties: Mapping[str, JsonSchema] = {},
    readOnly: bool = False,
    required: Sequence[str] = [],
    title: str = None,
    type: Union[JsonType, Sequence[JsonType]] = None,
    uniqueItems: bool = False,
    unevaluatedProperties: Union[bool, JsonSchema] = JsonSchema(),
    writeOnly: bool = False,
) -> JsonSchema:
    ...
